"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import uuid

from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('lambda_connection')


def _str_list_to_list(param, param_name):
    result = None
    if isinstance(param, list):
        result = param
    elif isinstance(param, str):
        result = [param]
    else:
        raise ValueError(
            '{} must be a str or a list of str.'.format(param_name))
    return result


@apply_methods_decorator(retry)
class LambdaConnection(object):
    """ Lambda connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('lambda', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Lambda connection.')

    def create_lambda(self, lambda_name, func_name,
                      role, s3_bucket, s3_key, runtime='python2.7', memory=128,
                      timeout=300, vpc_sub_nets=None, vpc_security_group=None,
                      env_vars=None, dl_target_arn=None, tracing_mode=None,
                      publish_version=False, layers=None):
        """ Create Lambda method.

        :type lambda_name: str
        :type func_name: str
        :param func_name: name of the entry point function
        :type role: str
        :param role: aws arn of role
        :type s3_bucket: str
        :type s3_key: str
        :type runtime: str
        :param memory: max limit of Lambda memory usage
        :param timeout: max limit of Lambda run time (300 - max value)
        :type vpc_sub_nets: list
        :type vpc_security_group: list
        :type env_vars: dict
        :param env_vars: {'string': 'string'}
        :type layers: list
        :return: response
        """
        layers = [] if layers is None else layers
        params = dict(FunctionName=lambda_name, Runtime=runtime,
                      Role=role, Handler=func_name,
                      Code={'S3Bucket': s3_bucket, 'S3Key': s3_key},
                      Description=' ', Timeout=timeout, MemorySize=memory,
                      Publish=publish_version, Layers=layers)
        if env_vars:
            params['Environment'] = {'Variables': env_vars}
        if vpc_sub_nets and vpc_security_group:
            params['VpcConfig'] = {
                'SubnetIds': vpc_sub_nets,
                'SecurityGroupIds': vpc_security_group
            }
        if dl_target_arn:
            params['DeadLetterConfig'] = {
                'TargetArn': dl_target_arn
            }
        if tracing_mode:
            params['TracingConfig'] = {
                'Mode': tracing_mode
            }
        return self.client.create_function(**params)

    def create_alias(self, function_name, name, version,
                     description=None, routing_config=None):
        """

        :param function_name: str
        :param name: str
        :param version: str
        :param description: str
        :param routing_config: dict str:int
        """
        params = dict(FunctionName=function_name, Name=name,
                      FunctionVersion=version)
        if description:
            params['Description'] = description
        if routing_config:
            params['RoutingConfig'] = routing_config

        return self.client.create_alias(**params)

    def get_alias(self, function_name, name):
        all_aliases = {}
        next_marker = 1  # to enter the loop
        while next_marker:
            req_param = {
                'FunctionName': function_name
            }
            if type(next_marker) == str:
                req_param['Marker'] = next_marker

            response = self.client.list_aliases(**req_param)

            all_aliases.update({item.get('Name'): item for item in
                                response.get('Aliases')})
            if all_aliases.get(name):
                return all_aliases.get(name)
            next_marker = response.get('NextMarker')

    def add_event_source(self, func_name, stream_arn, batch_size=15,
                         start_position=None):
        """ Create event source for Lambda.

        :type func_name: str
        :type stream_arn: str
        :param batch_size: max limit of Lambda event process in one time
        :param start_position: option for Lambda reading event mode
        :return: response
        """
        params = dict(EventSourceArn=stream_arn,
                      FunctionName=func_name,
                      Enabled=True,
                      BatchSize=batch_size)

        if start_position:
            params['StartingPosition'] = start_position

        response = self.client.create_event_source_mapping(**params)
        return response

    def lambdas_list(self):
        """ Get all existing Lambdas.

        :return: list(if exists)
        """
        lambdas = []
        response = self.client.list_functions(MaxItems=1000)
        if 'Functions' in response:
            lambdas.extend(response['Functions'])
        marker = response.get('NextMarker')
        while marker:
            response = self.client.list_functions(Marker=marker,
                                                  MaxItems=1000)
            if 'Functions' in response:
                lambdas.extend(response['Functions'])
            marker = response.get('NextMarker')
        return lambdas

    def triggers_list(self, lambda_name):
        """ Get all existing triggers.

        :return: list(if exists)
        """
        mappings = []
        response = self.client.list_event_source_mappings(
            FunctionName=lambda_name)
        token = response.get('NextMarker')
        mappings.extend(response.get('EventSourceMappings'))
        while token:
            response = self.client.list_event_source_mappings(Marker=token)
            token = response.get('NextMarker')
            mappings.extend(response.get('EventSourceMappings'))
        return mappings

    def versions_list(self, function_name):
        versions = []
        response = self.client.list_versions_by_function(
            FunctionName=function_name)
        if 'Versions' in response:
            versions.extend(response['Versions'])
        marker = response.get('NextMarker')
        while marker:
            response = self.client.list_versions_by_function(Marker=marker)
            if 'Versions' in response:
                versions.extend(response['Versions'])
            marker = response.get('NextMarker')
        return versions

    def delete_lambda(self, func_name):
        """ Delete Lambda.

        :param func_name: str
        """
        self.client.delete_function(FunctionName=func_name)

    def remove_trigger(self, lambda_name):
        """ Remove trigger by name. Trigger has the same name as Lambda.

        :type lambda_name: str
        """
        triggers = self.triggers_list(lambda_name)
        for trigger in triggers:
            trigger_name = trigger['FunctionArn'].split(':')[-1]
            if trigger_name == lambda_name:
                try:
                    self.client.delete_event_source_mapping(
                        UUID=trigger['UUID'])
                except ClientError:
                    _LOG.error('Failed to delete trigger.', exc_info=True)

    def remove_lambdas(self):
        """ Removes all specified lambdas.

        :type lambda_names: list
        """
        list_functions = self.lambdas_list()
        for each in list_functions:
            try:
                self.delete_lambda(each['FunctionName'])
            except ClientError:
                _LOG.error(
                    'Failed to delete lambda %s', each['FunctionName'],
                    exc_info=True)

    def delete_trigger(self, uuid):
        """ Delete event source stream.

        :param uuid: str
        """
        self.client.delete_event_source_mapping(UUID=uuid)

    def add_invocation_permission(self, name, principal, source_arn=None,
                                  statement_id=None):
        """ Add permission for something to be able invoke lambda.

        :type name: str
        :type source_arn: str
        :type principal: str
        :type statement_id: str
        """
        if not statement_id:
            statement_id = str(uuid.uuid1())
        params = dict(FunctionName=name, StatementId=statement_id,
                      Action='lambda:InvokeFunction', Principal=principal)
        if source_arn:
            params['SourceArn'] = source_arn
        self.client.add_permission(**params)

    def update_code_source(self, lambda_name, s3_bucket, s3_key,
                           publish_version):
        """ Update code source (s3 bucket + file link) for specified lambda.

        :type lambda_name: str
        :type s3_bucket: str
        :type s3_key: str
        :type publish_version: bool
        """
        self.client.update_function_code(FunctionName=lambda_name,
                                         S3Bucket=s3_bucket,
                                         S3Key=s3_key,
                                         Publish=publish_version)

    def update_event_source(self, lambda_name, batch_size):
        """ Update batch size of lambda event source stream.

        :type lambda_name: str
        :type batch_size: int
        """
        triggers = self.triggers_list(lambda_name)
        for trigger in triggers:
            trigger_name = trigger['FunctionArn'].split(':')[-1]
            if trigger_name == lambda_name:
                return self.client.update_event_source_mapping(
                    UUID=trigger['UUID'],
                    FunctionName=lambda_name,
                    Enabled=True,
                    BatchSize=batch_size
                )

    def get_function(self, lambda_name, qualifier=None):
        """ Get function info if it is exists,
        else - ResourceNotFoundException.

        :type lambda_name: name
        :param qualifier: Using this optional parameter to specify a function
        version or an alias name.
        :type qualifier: str
        """
        params = dict(FunctionName=lambda_name)
        if qualifier:
            params['Qualifier'] = qualifier
        try:
            return self.client.get_function(**params)
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_policy(self, lambda_name, qualifier=None):
        """ Returns the resource-based IAM policy for a function,
        version, or alias.

        :type lambda_name: name
        :param qualifier: Using this optional parameter to specify a function
        version or an alias name.
        :type qualifier: str
        """
        params = dict(FunctionName=lambda_name)
        if qualifier:
            params['Qualifier'] = qualifier
        try:
            return self.client.get_policy(**params)
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                pass  # valid exception
            else:
                raise e

    def invoke_lambda(self, lambda_name, invocation_type='Event',
                      log_type='Tail', client_context='', payload=b'',
                      qualifier=''):
        """

        :param lambda_name:
        :param invocation_type: 'Event' | 'RequestResponse' | 'DryRun'
        :param log_type: 'None' | 'Tail'
        :type client_context: str
        :type payload: b'bytes' | file
        :type qualifier: str
        :return:
        """
        params = dict(FunctionName=lambda_name,
                      InvocationType=invocation_type,
                      LogType=log_type,
                      ClientContext=client_context,
                      Payload=payload)
        if qualifier:
            params['Qualifier'] = qualifier
        return self.client.invoke(**params)

    def get_lambda_configuration(self, lambda_name, qualifier=None):
        params = dict(FunctionName=lambda_name)
        if qualifier:
            params['Qualifier'] = qualifier

        return self.client.get_function_configuration(**params)

    def update_lambda_configuration(self, lambda_name, role=None, handler=None,
                                    description=None, timeout=None,
                                    memory_size=None, vpc_sub_nets=None,
                                    vpc_security_group=None,
                                    env_vars=None, runtime=None,
                                    dead_letter_arn=None, kms_key_arn=None,
                                    layers=None):
        params = dict(FunctionName=lambda_name)
        if layers:
            params['Layers'] = layers
        if role:
            params['Role'] = role
        if handler:
            params['Handler'] = handler
        if description:
            params['Description'] = description
        if timeout:
            params['Timeout'] = timeout
        if memory_size:
            params['MemorySize'] = memory_size
        if vpc_sub_nets:
            vpc_sub_nets = _str_list_to_list(vpc_sub_nets, 'VPC_SUB_NETS')
        if vpc_security_group:
            vpc_sub_nets = _str_list_to_list(vpc_security_group,
                                             'VPC_SECURITY_GROUPS')
        if vpc_sub_nets and vpc_security_group:
            params['VpcConfig'] = {
                'SubnetIds': vpc_sub_nets,
                'SecurityGroupIds': vpc_security_group
            }
        if env_vars:
            params['Environment'] = {'Variables': env_vars}
        if runtime:
            params['Runtime'] = runtime
        if dead_letter_arn:
            params['DeadLetterConfig'] = {'TargetArn': dead_letter_arn}
        if kms_key_arn:
            params['KMSKeyArn'] = kms_key_arn
        return self.client.update_function_configuration(**params)

    def put_function_concurrency(self, function_name, concurrent_executions):
        return self.client.put_function_concurrency(
            FunctionName=function_name,
            ReservedConcurrentExecutions=concurrent_executions)

    def get_unresolved_concurrent_executions(self):
        return self.client.get_account_settings()['AccountLimit'][
            'UnreservedConcurrentExecutions']

    def publish_version(self, function_name, code_sha_256):
        return self.client.publish_version(
            FunctionName=function_name,
            CodeSha256=code_sha_256
        )

    def update_alias(self, function_name, alias_name, function_version):
        return self.client.update_alias(
            FunctionName=function_name,
            Name=alias_name,
            FunctionVersion=function_version
        )

    def create_layer(self, layer_name, s3_bucket, s3_key, runtimes,
                     description=None,
                     layer_license=None):
        kwargs = {'LayerName': layer_name, 'CompatibleRuntimes': runtimes,
                  'Content': {'S3Bucket': s3_bucket, 'S3Key': s3_key}}
        if description:
            kwargs['Description'] = description
        if layer_license:
            kwargs['LicenseInfo'] = layer_license
        return self.client.publish_layer_version(**kwargs)

    def get_lambda_layer_arn(self, name):
        lambda_layers = self.client.list_layers()
        for each in lambda_layers['Layers']:
            if each['LayerName'] == name:
                return each['LatestMatchingVersion']['LayerVersionArn']
        while lambda_layers.get('NextMarker'):
            lambda_layers = self.client.list_layers(
                Marker=lambda_layers.get('NextMarker'))
            for each in lambda_layers['Layers']:
                if each['LayerName'] == name:
                    return each['LatestMatchingVersion']['LayerVersionArn']

    def get_lambda_layer_by_arn(self, arn):
        return self.client.get_layer_version_by_arn(Arn=arn)

    def delete_layer(self, arn):
        version = arn.split(':')[len(arn.split(':')) - 1]
        arn = arn[:-len(version) - 1]
        return self.client.delete_layer_version(
            LayerName=arn,
            VersionNumber=int(version))

    def list_lambda_layer_versions(self, name, runtime=None):
        kwargs = {'LayerName': name}
        if runtime:
            kwargs['CompatibleRuntime'] = runtime
        response = self.client.list_layer_versions(**kwargs)
        versions = response['LayerVersions']

        while response.get('NextMarker'):
            kwargs['Marker'] = response['NextMarker']
            response = self.client.list_layer_versions(kwargs)
            versions.append(response['LayerVersions'])

        return versions

    def configure_provisioned_concurrency(self, name, qualifier,
                                          concurrent_executions):
        if type(concurrent_executions) is not int:
            raise AssertionError(
                f'Parameter `concurrent_executions` '
                f'must be type of int, but not {type(concurrent_executions)}')
        return self.client.put_provisioned_concurrency_config(
            FunctionName=name,
            Qualifier=qualifier,
            ProvisionedConcurrentExecutions=concurrent_executions
        )

    def get_provisioned_concurrency(self, name, qualifier):
        return self.client.get_provisioned_concurrency_config(
            FunctionName=name,
            Qualifier=qualifier
        )

    def describe_provisioned_concurrency_configs(self, name):
        configs = []
        response = self.client.list_provisioned_concurrency_configs(
            FunctionName=name
        )
        configs.extend(response.get('ProvisionedConcurrencyConfigs'))
        marker = response.get('NextMarker')
        while marker:
            response = self.client.list_provisioned_concurrency_configs(
                FunctionName=name,
                Marker=marker
            )
            configs.extend(response.get('ProvisionedConcurrencyConfigs'))
            marker = response.get('NextMarker')
        return configs

    def describe_function_concurrency(self, name):
        return self.client.get_function_concurrency(
            FunctionName=name
        ).get('ReservedConcurrentExecutions')

    def delete_function_concurrency_config(self, name):
        # client.delete_function_concurrency return 204 None in boto3 1.11.14
        self.client.delete_function_concurrency(
            FunctionName=name)

    def delete_provisioned_concurrency_config(self, name, qualifier):
        # client.delete_provisioned_concurrency_config return 204 None
        #   in boto3 1.11.14
        self.client.delete_provisioned_concurrency_config(
            FunctionName=name,
            Qualifier=qualifier
        )

    def list_function_versions(self, name):
        versions = []
        resp = self.client.list_versions_by_function(
            FunctionName=name,
            MaxItems=100
        )
        versions.extend(resp.get('Versions'))
        next_marker = resp.get('NextMarker')
        while next_marker:
            self.client.list_versions_by_function(
                FunctionName=name,
                MaxItems=100,
                Marker=next_marker
            )
            versions.extend(resp.get('Versions'))
            next_marker = resp.get('NextMarker')
        return versions



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('core.resources.sqs_resource')

AVAILABLE_REGIONS = ['us-east-2', 'us-east-1', 'us-west-1', 'us-west-2',
                     'ap-south-1', 'ap-northeast-2', 'ap-southeast-1',
                     'ap-southeast-2', 'ap-northeast-1', 'ca-central-1',
                     'eu-central-1', 'eu-west-2', 'sa-east-1', 'eu-west-1']
FIFO_REGIONS = ['us-east-1', 'us-east-2', 'us-west-2', 'eu-west-1']


class SqsResource(BaseResource):

    def __init__(self, sqs_conn_builder, region, account_id) -> None:
        self.sqs_conn_builder = sqs_conn_builder
        self.region = region
        self.account_id = account_id

    def create_sqs_queue(self, args):
        return self.create_pool(self._create_sqs_queue_from_meta, args)

    def describe_queue(self, queue_url, name, meta, resource_name, region):
        response = self.sqs_conn_builder(region).get_queue_attributes(
            queue_url)
        arn = self._build_queue_arn(resource_name=resource_name, region=region)
        return {arn: build_description_obj(response, name, meta)}

    def describe_queue_from_meta(self, name, meta):
        region = meta.get('region', self.region)
        is_fifo = meta.get('fifo_queue', False)
        resource_name = self._build_resource_name(is_fifo, name)
        queue_url = self.sqs_conn_builder(region).get_queue_url(
            resource_name, self.account_id)
        if not queue_url:
            return {}
        response = self.sqs_conn_builder(region).get_queue_attributes(
            queue_url)
        return {
            self._build_queue_arn(resource_name,
                                  region): build_description_obj(
                response, name, meta)
        }

    def remove_queues(self, args):
        self.create_pool(self._remove_queue, args)
        # wait to remove all queues
        if args:
            time.sleep(60)

    @unpack_kwargs
    def _remove_queue(self, arn, config):
        region = arn.split(':')[3]
        queue_name = config['resource_name']
        resource_meta = config['resource_meta']
        try:
            is_fifo = resource_meta.get('fifo_queue', False)
            resource_name = self._build_resource_name(is_fifo, queue_name)
            queue_url = self.sqs_conn_builder(region).get_queue_url(
                resource_name,
                self.account_id)
            if queue_url:
                self.sqs_conn_builder(region).delete_queue(queue_url)
                _LOG.info('SQS queue %s was removed.', queue_name)
            else:
                _LOG.warn('SQS queue %s is not found', queue_name)
        except ClientError as e:
            exception_type = e.response['Error']['Code']
            if exception_type == 'ResourceNotFoundException':
                _LOG.warn('SQS queue %s is not found', queue_name)
            else:
                raise e

    @unpack_kwargs
    def _create_sqs_queue_from_meta(self, name, meta):
        region = meta.get('region', self.region)
        is_fifo = meta.get('fifo_queue', False)
        resource_name = self._build_resource_name(is_fifo, name)
        queue_url = self.sqs_conn_builder(region).get_queue_url(resource_name,
                                                                self.account_id)
        if queue_url:
            _LOG.warn('SQS queue %s exists.', name)
            return self.describe_queue(queue_url, name, meta, resource_name,
                                       region)
        delay_sec = meta.get('delay_seconds')
        max_mes_size = meta.get('maximum_message_size')
        mes_ret_period = meta.get('message_retention_period')
        policy = meta.get('policy')
        recieve_mes_wait_sec = meta.get('receive_message_wait_time_seconds')
        redrive_policy = meta.get('redrive_policy')
        vis_timeout = meta.get('visibility_timeout')
        kms_master_key_id = meta.get('kms_master_key_id')
        kms_data_reuse_period = meta.get('kms_data_key_reuse_period_seconds')
        if is_fifo and region not in FIFO_REGIONS:
            raise AssertionError('FIFO queue is not available in {0}.',
                                 region)
        content_deduplication = meta.get('content_based_deduplication')
        params = dict(queue_name=resource_name,
                      delay_seconds=delay_sec,
                      maximum_message_size=max_mes_size,
                      message_retention_period=mes_ret_period,
                      policy=policy,
                      receive_message_wait_time_seconds=recieve_mes_wait_sec,
                      redrive_policy=redrive_policy,
                      visibility_timeout=vis_timeout,
                      kms_master_key_id=kms_master_key_id,
                      kms_data_key_reuse_period_seconds=kms_data_reuse_period,
                      fifo_queue=is_fifo,
                      content_based_deduplication=content_deduplication)
        queue_url = self.sqs_conn_builder(region).create_queue(**params)[
            'QueueUrl']
        _LOG.info('Created SQS queue %s.', name)
        return self.describe_queue(queue_url, name, meta, resource_name,
                                   region)

    @staticmethod
    def _build_resource_name(is_fifo, name):
        resource_name = name
        if is_fifo:
            resource_name += '.fifo'
        return resource_name

    def _build_queue_arn(self, resource_name, region):
        arn = 'arn:aws:sqs:{0}:{1}:{2}'.format(region, self.account_id,
                                               resource_name)
        return arn



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import json
import os
import shutil
import threading
from concurrent.futures.thread import ThreadPoolExecutor

from syndicate.commons.log_helper import get_logger
from syndicate.core.build.helper import build_py_package_name, zip_dir
from syndicate.core.constants import (LAMBDA_CONFIG_FILE_NAME,
                                      NODE_REQ_FILE_NAME)
from syndicate.core.helper import build_path, unpack_kwargs, execute_command
from syndicate.core.resources.helper import validate_params

_LOG = get_logger('nodejs_runtime_assembler')


def assemble_node_lambdas(project_path, bundles_dir):
    from syndicate.core import CONFIG
    project_abs_path = build_path(CONFIG.project_path, project_path)

    _LOG.info('Going to package lambdas starting by path {0}'.format(
        project_abs_path))
    executor = ThreadPoolExecutor(max_workers=5)
    futures = []
    for root, sub_dirs, files in os.walk(project_abs_path):
        for item in files:
            if item.endswith(LAMBDA_CONFIG_FILE_NAME):
                _LOG.info('Going to build artifact in: {0}'.format(root))
                arg = {
                    'item': item,
                    'root': root,
                    'target_folder': bundles_dir
                }
                futures.append(executor.submit(_build_node_artifact, arg))
    for future in concurrent.futures.as_completed(futures):
        _LOG.info(future.result())


@unpack_kwargs
def _build_node_artifact(item, root, target_folder):
    _check_npm_is_installed()
    _LOG.debug('Building artifact in {0}'.format(target_folder))
    lambda_config_dict = json.load(open(build_path(root, item)))
    _LOG.debug('Root path: {}'.format(root))
    req_params = ['lambda_path', 'name', 'version']
    validate_params(root, lambda_config_dict, req_params)
    lambda_name = lambda_config_dict['name']
    lambda_version = lambda_config_dict['version']
    artifact_name = lambda_name + '-' + lambda_version
    # create folder to store artifacts
    artifact_path = build_path(target_folder, artifact_name)
    _LOG.debug('Artifacts path: {0}'.format(artifact_path))
    if not os.path.exists(artifact_path):
        os.makedirs(artifact_path)
    _LOG.debug('Folders are created')
    # getting file content
    req_path = build_path(root, NODE_REQ_FILE_NAME)
    try:
        if os.path.exists(req_path):
            command = 'npm install --prefix {0}'.format(root)
            execute_command(command=command)
            _LOG.debug('3-rd party dependencies were installed successfully')

        package_name = build_py_package_name(lambda_name, lambda_version)
        zip_dir(root, build_path(target_folder, package_name))
        lock = threading.RLock()
        lock.acquire()
        try:
            # remove unused folder/files
            node_modules_path = os.path.join(root, 'node_modules')
            if os.path.exists(node_modules_path):
                shutil.rmtree(node_modules_path)
            # todo Investigate deleting package_lock file
            # shutil.rmtree(os.path.join(root, 'package_lock.json'))
            shutil.rmtree(artifact_path)
        except FileNotFoundError as e:
            _LOG.exception('Error occurred while temp files removing.')
        finally:
            lock.release()
        return 'Lambda package {0} was created successfully'.format(
            package_name)
    except Exception:
        _LOG.exception(
            'Error occurred during the \'{0}\' lambda deployment package '
            'assembling'.format(lambda_name))
        return 'Error occurred during the \'{0}\' lambda deployment package ' \
               'assembling'.format(lambda_name)


def _check_npm_is_installed():
    import subprocess
    result = subprocess.call('npm -v', shell=True)
    if result:
        raise AssertionError(
            'NPM is not installed. There is no ability to build '
            'NodeJS bundle. Please, install npm are retry to build bundle.')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             validate_params)

_LOG = get_logger('syndicate.core.resources.api_gateway_resource')

SUPPORTED_METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'HEAD']
_CORS_HEADER_NAME = 'Access-Control-Allow-Origin'
_CORS_HEADER_VALUE = "'*'"


class ApiGatewayResource(BaseResource):

    def __init__(self, apigw_conn, lambda_res, account_id, region) -> None:
        self.connection = apigw_conn
        self.lambda_res = lambda_res
        self.account_id = account_id
        self.region = region

    def api_resource_identifier(self, name, output=None):
        if output:
            # api currently is not located in different regions
            # process only first object
            api_output = list(output.items())[0][1]
            # find id from the output
            return api_output['description']['id']
        # if output is not provided - try to get API by name
        # cause there is no another option
        return self.connection.get_api_id(name)

    def create_api_gateway(self, args):
        """ Create api gateway in pool in sub processes.
    
        :type args: list
        """
        return self.create_pool(self._create_api_gateway_from_meta, args, 1)

    def api_gateway_update_processor(self, args):
        return self.create_pool(self._create_or_update_api_gateway, args, 1)

    @unpack_kwargs
    def _create_or_update_api_gateway(self, name, meta,
                                      current_configurations):
        if current_configurations:
            # api currently is not located in different regions
            # process only first object
            api_output = list(current_configurations.items())[0][1]
            # find id from the output
            api_id = api_output['description']['id']
            # check that api does not exist
            api_response = self.connection.get_api(api_id)
            if api_response:
                # find all existing resources
                existing_resources = api_output['description']['resources']
                existing_paths = [i['path'] for i in existing_resources]
                meta_api_resources = meta['resources']
                api_resp = meta.get('api_method_responses')
                api_integration_resp = meta.get(
                    'api_method_integration_responses')
                api_resources = {}
                for resource_path, resource_meta in meta_api_resources.items():
                    if resource_path not in existing_paths:
                        api_resources[resource_path] = resource_meta
                if api_resources:
                    _LOG.debug(
                        'Going to continue deploy API Gateway {0} ...'.format(
                            api_id))
                    args = self.__prepare_api_resources_args(
                        api_id=api_id,
                        api_resources=api_resources,
                        api_resp=api_resp,
                        api_integration_resp=api_integration_resp)
                    self.create_pool(self._create_resource_from_metadata,
                                     args, 1)
                    # add headers
                    # waiter b4 customization
                    time.sleep(10)
                    _LOG.debug(
                        'Customizing API Gateway {0} responses...'.format(
                            api_id))
                else:
                    # all resources created, but need to override
                    api_resources = meta_api_resources
                # _customize_gateway_responses call is commented due to
                # botocore InternalFailure while performing the call.
                # will be fixed later
                # _customize_gateway_responses(api_id)
                # deploy api
                _LOG.debug('Deploying API Gateway {0} ...'.format(api_id))
                self.__deploy_api_gateway(api_id, meta, api_resources)
                return self.describe_api_resources(api_id=api_id, meta=meta,
                                                   name=name)
            else:
                # api does not exist, so create a new
                return self._create_api_gateway_from_meta(
                    {'name': name, 'meta': meta})
        else:
            # object is not present, so just create a new api
            return self._create_api_gateway_from_meta(
                {'name': name, 'meta': meta})

    def _escape_path(self, parameter):
        index = parameter.find('/', 0)
        if index == -1:
            return parameter
        parameter = parameter[:index] + '~1' + parameter[index + 1:]
        return self._escape_path(parameter)

    def configure_cache(self, api_id, stage_name, api_resources):
        for resource_path, resource_meta in api_resources.items():
            for method_name, method_meta in resource_meta.items():
                if method_name in SUPPORTED_METHODS:
                    cache_configuration = method_meta.get(
                        'cache_configuration')
                    if not cache_configuration:
                        continue
                    cache_ttl_setting = cache_configuration.get(
                        'cache_ttl_sec')
                    if cache_ttl_setting:
                        _LOG.info(
                            'Configuring cache for {0}; TTL: {1}'.format(
                                resource_path, cache_ttl_setting))
                        escaped_resource = self._escape_path(resource_path)
                        self.connection.update_configuration(
                            rest_api_id=api_id,
                            stage_name=stage_name,
                            patch_operations=[
                                {
                                    'op': 'replace',
                                    'path': '/{0}/{1}/caching/ttlInSeconds'.format(
                                        escaped_resource,
                                        method_name),
                                    'value': str(cache_ttl_setting),
                                },
                                {
                                    'op': 'replace',
                                    'path': '/{0}/{1}/caching/enabled'.format(
                                        escaped_resource,
                                        method_name),
                                    'value': 'True',
                                }
                            ]
                        )
                        _LOG.info(
                            'Cache for {0} was configured'.format(
                                resource_path))

    @unpack_kwargs
    def _create_api_gateway_from_meta(self, name, meta):
        """ Create API Gateway with all specified meta.
    
        :type name: str
        :type meta: dict
        """
        required_parameters = ['resources', 'deploy_stage']
        validate_params(name, meta, required_parameters)

        api_resources = meta['resources']

        api_item = self.connection.create_rest_api(name)
        api_id = api_item['id']

        # deploy authorizers
        authorizers = meta.get('authorizers', {})
        for key, val in authorizers.items():
            lambda_version = val.get('lambda_version')
            lambda_name = val.get('lambda_name')
            lambda_alias = val.get('lambda_alias')
            lambda_arn = self.lambda_res. \
                resolve_lambda_arn_by_version_and_alias(lambda_name,
                                                        lambda_version,
                                                        lambda_alias)
            uri = 'arn:aws:apigateway:{0}:lambda:path/2015-03-31/functions/{1}/invocations'.format(
                self.region, lambda_arn)
            self.connection.create_authorizer(api_id=api_id, name=key,
                                              type=val['type'],
                                              authorizer_uri=uri,
                                              identity_source=val.get(
                                                  'identity_source'),
                                              ttl=val.get('ttl'))

            self.lambda_res.add_invocation_permission(
                statement_id=api_id,
                name=lambda_arn,
                principal='apigateway.amazonaws.com')
        if api_resources:
            api_resp = meta.get('api_method_responses')
            api_integration_resp = meta.get('api_method_integration_responses')
            args = self.__prepare_api_resources_args(api_id, api_resources,
                                                     api_resp,
                                                     api_integration_resp)
            self.create_pool(self._create_resource_from_metadata, args, 1)
        else:
            _LOG.info('There is no resources in %s API Gateway description.',
                      name)
        # add headers
        # waiter b4 customization
        time.sleep(10)
        _LOG.debug('Customizing API Gateway responses...')
        # _customize_gateway_responses call is commented due to botocore
        # InternalFailure while performing the call. will be fixed later
        # _customize_gateway_responses(api_id)
        # deploy api
        self.__deploy_api_gateway(api_id, meta, api_resources)
        return self.describe_api_resources(api_id=api_id, meta=meta, name=name)

    def __deploy_api_gateway(self, api_id, meta, api_resources):
        deploy_stage = meta['deploy_stage']
        cache_cluster_configuration = meta.get('cluster_cache_configuration')
        root_cache_enabled = cache_cluster_configuration.get(
            'cache_enabled') if cache_cluster_configuration else None
        cache_size = cache_cluster_configuration.get(
            'cache_size') if cache_cluster_configuration else None
        self.connection.deploy_api(api_id, stage_name=deploy_stage,
                                   cache_cluster_enabled=root_cache_enabled,
                                   cache_cluster_size=str(
                                       cache_size) if cache_size else None)
        # configure caching
        if root_cache_enabled:
            _LOG.debug('Cluster cache configuration found:{0}'.format(
                cache_cluster_configuration))
            # set default ttl for root endpoint
            cluster_cache_ttl_sec = cache_cluster_configuration.get(
                'cache_ttl_sec')
            self.connection.update_configuration(
                rest_api_id=api_id,
                stage_name=deploy_stage,
                patch_operations=[
                    {
                        'op': 'replace',
                        'path': '/*/*/caching/ttlInSeconds',
                        'value': str(cluster_cache_ttl_sec),
                    }
                ]
            )
            # customize cache settings for endpoints
            self.configure_cache(api_id, deploy_stage, api_resources)

    def __prepare_api_resources_args(self, api_id, api_resources,
                                     api_resp=None,
                                     api_integration_resp=None):
        # describe authorizers and create a mapping
        authorizers = self.connection.get_authorizers(api_id)
        authorizers_mapping = {x['name']: x['id'] for x in authorizers}
        args = []
        for each in api_resources:
            resource_meta = api_resources[each]
            _LOG.info('Creating resource %s ...', each)
            if each.startswith('/'):
                resource_id = self.connection.get_resource_id(api_id, each)
                if resource_id:
                    _LOG.info('Resource %s exists.', each)
                    enable_cors = resource_meta.get('enable_cors')
                    self._check_existing_methods(
                        api_id=api_id, resource_id=resource_id,
                        resource_path=each,
                        resource_meta=resource_meta,
                        enable_cors=enable_cors,
                        authorizers_mapping=authorizers_mapping,
                        api_resp=api_resp,
                        api_integration_resp=api_integration_resp)
                else:
                    args.append({
                        'api_id': api_id,
                        'resource_path': each,
                        'resource_meta': resource_meta,
                        'authorizers_mapping': authorizers_mapping
                    })
            else:
                raise AssertionError(
                    "API resource must starts with '/', but found %s", each)
        return args

    def describe_api_resources(self, name, meta, api_id=None):
        if not api_id:
            api = self.connection.get_api_by_name(name)
            if not api:
                return
            api_id = api['id']

        response = self.connection.get_api(api_id)
        if not response:
            return
        response['resources'] = self.connection.get_resources(api_id)
        _LOG.info('Created %s API Gateway.', name)
        arn = 'arn:aws:apigateway:{0}::/restapis/{1}'.format(self.region,
                                                             api_id)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def _check_existing_methods(self, api_id, resource_id, resource_path,
                                resource_meta,
                                enable_cors, authorizers_mapping,
                                api_resp=None,
                                api_integration_resp=None):
        """ Check if all specified methods exist and create some if not.
    
        :type api_id: str
        :type resource_id: str
        :type resource_meta: dict
        :type enable_cors: bool or None
        :type:
        """
        for method in resource_meta:
            if method == 'enable_cors':
                continue
            if self.connection.get_method(api_id, resource_id, method):
                _LOG.info('Method %s exists.', method)
                continue
            else:
                _LOG.info('Creating method %s for resource %s...',
                          method, resource_id)
                self._create_method_from_metadata(
                    api_id=api_id,
                    resource_id=resource_id,
                    resource_path=resource_path,
                    method=method,
                    method_meta=resource_meta[method],
                    authorizers_mapping=authorizers_mapping,
                    api_resp=api_resp,
                    api_integration_resp=api_integration_resp,
                    enable_cors=enable_cors)
            if enable_cors and not self.connection.get_method(api_id,
                                                              resource_id,
                                                              'OPTIONS'):
                _LOG.info('Enabling CORS for resource %s...', resource_id)
                self.connection.enable_cors_for_resource(api_id, resource_id)

    @unpack_kwargs
    def _create_resource_from_metadata(self, api_id, resource_path,
                                       resource_meta,
                                       authorizers_mapping):
        self.connection.create_resource(api_id, resource_path)
        _LOG.info('Resource %s created.', resource_path)
        resource_id = self.connection.get_resource_id(api_id, resource_path)
        enable_cors = resource_meta.get('enable_cors')
        for method in resource_meta:
            try:
                if method == 'enable_cors' or method not in SUPPORTED_METHODS:
                    continue

                method_meta = resource_meta[method]
                _LOG.info('Creating method %s for resource %s...',
                          method, resource_path)
                self._create_method_from_metadata(
                    api_id=api_id,
                    resource_id=resource_id,
                    resource_path=resource_path,
                    method=method,
                    method_meta=method_meta,
                    enable_cors=enable_cors,
                    authorizers_mapping=authorizers_mapping)
            except Exception as e:
                _LOG.error('Resource: {0}, method {1}.'
                           .format(resource_path, method), exc_info=True)
                raise e
            _LOG.info('Method %s for resource %s created.', method,
                      resource_path)
        # create enable cors only after all methods in resource created
        if enable_cors:
            self.connection.enable_cors_for_resource(api_id, resource_id)
            _LOG.info('CORS enabled for resource %s', resource_path)

    def _create_method_from_metadata(self, api_id, resource_id, resource_path,
                                     method,
                                     method_meta, authorizers_mapping,
                                     enable_cors=False, api_resp=None,
                                     api_integration_resp=None):
        # init responses for method
        method_responses = method_meta.get("responses")
        if method_responses:
            resp = method_responses
        elif api_resp:
            resp = api_resp
        else:
            resp = []

        # init integration responses for method
        integration_method_responses = method_meta.get("integration_responses")
        if integration_method_responses:
            integr_resp = integration_method_responses
        elif api_resp:
            integr_resp = api_integration_resp
        else:
            integr_resp = []

        # resolve authorizer if needed
        authorization_type = method_meta.get('authorization_type')
        if authorization_type not in ['NONE', 'AWS_IAM']:
            # type is authorizer, so add id to meta
            authorizer_id = authorizers_mapping.get(authorization_type)
            if not authorizer_id:
                raise AssertionError(
                    'Authorizer {0} does not exist'.format(authorization_type))
            method_meta['authorizer_id'] = authorizer_id
            authorization_type = 'CUSTOM'

        self.connection.create_method(
            api_id, resource_id, method,
            authorization_type=authorization_type,
            authorizer_id=method_meta.get('authorizer_id'),
            api_key_required=method_meta.get('api_key_required'),
            request_parameters=method_meta.get('method_request_parameters'),
            request_models=method_meta.get('method_request_models'))
        # second step: create integration
        integration_type = method_meta.get('integration_type')
        # set up integration - lambda or aws service
        body_template = method_meta.get('integration_request_body_template')
        passthrough_behavior = method_meta.get(
            'integration_passthrough_behavior')
        # TODO split to map - func implementation
        if integration_type:
            if integration_type == 'lambda':
                lambda_name = method_meta['lambda_name']
                # alias has a higher priority than version in arn resolving
                lambda_version = method_meta.get('lambda_version')
                lambda_alias = method_meta.get('lambda_alias')
                lambda_arn = self.lambda_res. \
                    resolve_lambda_arn_by_version_and_alias(lambda_name,
                                                            lambda_version,
                                                            lambda_alias)
                enable_proxy = method_meta.get('enable_proxy')
                cache_configuration = method_meta.get('cache_configuration')
                cache_key_parameters = cache_configuration.get(
                    'cache_key_parameters') if cache_configuration else None
                self.connection.create_lambda_integration(
                    lambda_arn, api_id, resource_id, method, body_template,
                    passthrough_behavior, method_meta.get('lambda_region'),
                    enable_proxy=enable_proxy,
                    cache_key_parameters=cache_key_parameters)
                # add permissions to invoke
                api_source_arn = f"arn:aws:execute-api:{self.region}:" \
                    f"{self.account_id}:{api_id}/*/{method}{resource_path}"
                self.lambda_res.add_invocation_permission(
                    statement_id=(api_id + '-' + method + '-' +
                                  resource_path[1:].replace('/', '_').
                                  replace('{', '').
                                  replace('}', '')),
                    name=lambda_arn,
                    principal='apigateway.amazonaws.com',
                    source_arn=api_source_arn)
            elif integration_type == 'service':
                uri = method_meta.get('uri')
                role = method_meta.get('role')
                integration_method = method_meta.get('integration_method')
                self.connection.create_service_integration(self.account_id,
                                                           api_id,
                                                           resource_id,
                                                           method,
                                                           integration_method,
                                                           role, uri,
                                                           body_template,
                                                           passthrough_behavior)
            elif integration_type == 'mock':
                self.connection.create_mock_integration(api_id, resource_id,
                                                        method,
                                                        body_template,
                                                        passthrough_behavior)
            elif integration_type == 'http':
                integration_method = method_meta.get('integration_method')
                uri = method_meta.get('uri')
                enable_proxy = method_meta.get('enable_proxy')
                self.connection.create_http_integration(api_id, resource_id,
                                                        method,
                                                        integration_method,
                                                        uri,
                                                        body_template,
                                                        passthrough_behavior,
                                                        enable_proxy)
            else:
                raise AssertionError('%s integration type does not exist.',
                                     integration_type)
        # third step: setup method responses
        if resp:
            for response in resp:
                self.connection.create_method_response(
                    api_id, resource_id, method, response.get('status_code'),
                    response.get('response_parameters'),
                    response.get('response_models'), enable_cors)
        else:
            self.connection.create_method_response(
                api_id, resource_id, method, enable_cors=enable_cors)
        # fourth step: setup integration responses
        if integr_resp:
            for each in integr_resp:
                self.connection.create_integration_response(
                    api_id, resource_id, method, each.get('status_code'),
                    each.get('error_regex'),
                    each.get('response_parameters'),
                    each.get('response_templates'), enable_cors)
        else:
            self.connection.create_integration_response(
                api_id, resource_id, method, enable_cors=enable_cors)

    def _check_existing_methods(self, api_id, resource_id, resource_path,
                                resource_meta,
                                enable_cors, authorizers_mapping,
                                api_resp=None,
                                api_integration_resp=None):
        """ Check if all specified methods exist and create some if not.
    
        :type api_id: str
        :type resource_id: str
        :type resource_meta: dict
        :type enable_cors: bool or None
        :type:
        """
        for method in resource_meta:
            if method == 'enable_cors':
                continue
            if self.connection.get_method(api_id, resource_id, method):
                _LOG.info('Method %s exists.', method)
                continue
            else:
                _LOG.info('Creating method %s for resource %s...',
                          method, resource_id)
                self._create_method_from_metadata(
                    api_id=api_id,
                    resource_id=resource_id,
                    resource_path=resource_path,
                    method=method,
                    method_meta=resource_meta[method],
                    authorizers_mapping=authorizers_mapping,
                    api_resp=api_resp,
                    api_integration_resp=api_integration_resp,
                    enable_cors=enable_cors)
            if enable_cors and not self.connection.get_method(api_id,
                                                              resource_id,
                                                              'OPTIONS'):
                _LOG.info('Enabling CORS for resource %s...', resource_id)
                self.connection.enable_cors_for_resource(api_id, resource_id)

        @unpack_kwargs
        def _create_resource_from_metadata(self, api_id, resource_path,
                                           resource_meta,
                                           authorizers_mapping):
            self.connection.create_resource(api_id, resource_path)
            _LOG.info('Resource %s created.', resource_path)
            resource_id = self.connection.get_resource_id(api_id,
                                                          resource_path)
            enable_cors = resource_meta.get('enable_cors')
            for method in resource_meta:
                try:
                    if method == 'enable_cors' or method not in SUPPORTED_METHODS:
                        continue

                    method_meta = resource_meta[method]
                    _LOG.info('Creating method %s for resource %s...',
                              method, resource_path)
                    self._create_method_from_metadata(
                        api_id=api_id,
                        resource_id=resource_id,
                        resource_path=resource_path,
                        method=method,
                        method_meta=method_meta,
                        enable_cors=enable_cors,
                        authorizers_mapping=authorizers_mapping)
                except Exception as e:
                    _LOG.error('Resource: {0}, method {1}.'
                               .format(resource_path, method), exc_info=True)
                    raise e
                _LOG.info('Method %s for resource %s created.', method,
                          resource_path)
            # create enable cors only after all methods in resource created
            if enable_cors:
                self.connection.enable_cors_for_resource(api_id, resource_id)
                _LOG.info('CORS enabled for resource %s', resource_path)

    def _customize_gateway_responses(self, api_id):
        responses = self.connection.get_gateway_responses(api_id)
        response_types = [r['responseType'] for r in responses]
        for response_type in response_types:
            time.sleep(10)
            self.connection.add_header_to_gateway_response(api_id,
                                                           response_type,
                                                           _CORS_HEADER_NAME,
                                                           _CORS_HEADER_VALUE)

    def remove_api_gateways(self, args):
        for arg in args:
            self._remove_api_gateway(**arg)
            # wait for success deletion
            time.sleep(60)

    def _remove_api_gateway(self, arn, config):
        api_id = config['description']['id']
        try:
            self.connection.remove_api(api_id)
            _LOG.info('API Gateway %s was removed.', api_id)
        except ClientError as e:
            if e.response['Error']['Code'] == 'NotFoundException':
                _LOG.warn('API Gateway %s is not found', api_id)
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection import ConnectionProvider
from syndicate.connection.sts_connection import STSConnection
from syndicate.core.conf.processor import ConfigHolder
from syndicate.core.resources.processors_mapping import ProcessorFacade
from syndicate.core.resources.resources_provider import ResourceProvider

_LOG = get_logger('deployment.__init__')

SESSION_TOKEN = 'aws_session_token'
SECRET_KEY = 'aws_secret_access_key'
ACCESS_KEY = 'aws_access_key_id'


def exception_handler(exception_type, exception, traceback):
    print(exception)


# sys.excepthook = exception_handler

# CONF VARS ===================================================================
CONF_PATH = os.environ.get('SDCT_CONF')
CONFIG = None
CONN = None
CREDENTIALS = None
RESOURCES_PROVIDER = None
PROCESSOR_FACADE = None


def _ready_to_assume():
    return CONFIG.access_role and CONFIG.aws_access_key_id and \
           CONFIG.aws_secret_access_key


def _ready_to_use_creds():
    return not CONFIG.access_role and CONFIG.aws_access_key_id and \
           CONFIG.aws_secret_access_key


def initialize_connection():
    global CONFIG
    global CONN
    global CONF_PATH
    global CREDENTIALS
    global RESOURCES_PROVIDER
    global PROCESSOR_FACADE

    CONFIG = ConfigHolder(CONF_PATH)
    sts = STSConnection(CONFIG.region, CONFIG.aws_access_key_id,
                        CONFIG.aws_secret_access_key)
    try:
        CREDENTIALS = {
            'region': CONFIG.region
        }
        if _ready_to_assume():
            _LOG.debug('Starting to assume role ...')
            # get CREDENTIALS for N hours
            temp_credentials = sts.get_temp_credentials(CONFIG.access_role,
                                                        CONFIG.account_id,
                                                        CONFIG.session_duration)
            _LOG.debug(f'Role {CONFIG.access_role} is assumed successfully'
                       f'for {CONFIG.session_duration} seconds')
            CREDENTIALS[ACCESS_KEY] = temp_credentials['AccessKeyId']
            CREDENTIALS[SECRET_KEY] = temp_credentials['SecretAccessKey']
            CREDENTIALS[SESSION_TOKEN] = temp_credentials['SessionToken']
        elif _ready_to_use_creds():
            _LOG.debug('Credentials access')
            CREDENTIALS[ACCESS_KEY] = CONFIG.aws_access_key_id
            CREDENTIALS[SECRET_KEY] = CONFIG.aws_secret_access_key
        CONN = ConnectionProvider(CREDENTIALS)
        RESOURCES_PROVIDER = ResourceProvider(config=CONFIG,
                                              credentials=CREDENTIALS,
                                              sts_conn=sts)
        PROCESSOR_FACADE = ProcessorFacade(
            resources_provider=RESOURCES_PROVIDER)
        _LOG.debug('aws-syndicate has been initialized')
    except ClientError:
        raise AssertionError('Cannot assume {0} role. '
                             'Please verify that you have configured '
                             'the role correctly.'.format(CONFIG.access_role))



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.conf.processor import ALL_REGIONS
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             check_region_available,
                                             create_args_for_multi_region,
                                             validate_params)

_LOG = get_logger('core.resources.sns_resource')


class SnsResource(BaseResource):

    def __init__(self, conn_provider, region) -> None:
        self.connection_provider = conn_provider
        self.region = region

    def describe_sns(self, name, meta, region, arn=None):
        if not arn:
            arn = self.connection_provider.sns(region).get_topic_arn(name)
        response = self.connection_provider.sns(region).get_topic_attributes(
            arn)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def describe_sns_from_meta(self, name, meta):
        new_region_args = create_args_for_multi_region(
            [
                {
                    'name': name,
                    'meta': meta
                }
            ],
            ALL_REGIONS)
        responses = []
        for arg in new_region_args:
            region = arg['region']
            topic_arn = self.connection_provider.sns(region).get_topic_arn(
                name)
            if not topic_arn:
                continue
            response = self.connection_provider.sns(
                region).get_topic_attributes(
                topic_arn)
            if response:
                responses.append({'arn': topic_arn, 'response': response})
        description = []
        for topic in responses:
            description.append({
                topic['arn']: build_description_obj(
                    topic['response'], name, meta)
            })
        return description

    def describe_sns_application(self, name, meta, region, arn=None):
        if not arn:
            arn = self.connection_provider.sns(
                region).get_platform_application(name)
        response = self.connection_provider.sns(
            region).get_platform_application_attributes(arn)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def describe_sns_application_from_meta(self, name, meta):
        new_region_args = create_args_for_multi_region(
            [
                {
                    'name': name,
                    'meta': meta
                }
            ],
            ALL_REGIONS)
        responses = []
        for arg in new_region_args:
            region = arg['region']
            app_arn = self.connection_provider.sns(
                region).get_platform_application(
                name)
            if not app_arn:
                continue
            response = self.connection_provider.sns(
                region).get_platform_application_attributes(
                app_arn)
            if response:
                responses.append({'arn': app_arn, 'response': response})
        description = []
        for topic in responses:
            description.append({
                topic['arn']: build_description_obj(
                    topic['response'], name, meta)
            })
        return description

    def create_sns_topic(self, args):
        """ Create sns topic from meta in region/regions.

        :type args: list
        """
        new_region_args = create_args_for_multi_region(args, ALL_REGIONS)
        return self.create_pool(self._create_sns_topic_from_meta,
                                new_region_args)

    def create_sns_application(self, args):
        """ Create sns application from meta in region/regions.

        :type args: list
        """
        new_region_args = create_args_for_multi_region(args, ALL_REGIONS)
        return self.create_pool(self._create_platform_application_from_meta,
                                new_region_args)

    @unpack_kwargs
    def _create_sns_topic_from_meta(self, name, meta, region):
        arn = self.connection_provider.sns(region).get_topic_arn(name)
        if arn:
            _LOG.warn(
                '{0} sns topic exists in region {1}.'.format(name, region))
            return self.describe_sns(name=name, meta=meta, region=region,
                                     arn=arn)
        arn = self.connection_provider.sns(region).create_topic(name)
        event_sources = meta.get('event_sources')
        if event_sources:
            for trigger_meta in event_sources:
                trigger_type = trigger_meta['resource_type']
                func = self.CREATE_TRIGGER[trigger_type]
                func(name, trigger_meta, region)
        _LOG.info('SNS topic %s in region %s created.', name, region)
        return self.describe_sns(name=name, meta=meta, region=region, arn=arn)

    def _subscribe_lambda_to_sns_topic(self, lambda_arn, topic_name, region):
        topic_arn = self.connection_provider.sns(region).subscribe(lambda_arn,
                                                                   topic_name,
                                                                   'lambda')
        try:
            self.connection_provider.lambda_conn().add_invocation_permission(
                lambda_arn,
                'sns.amazonaws.com',
                source_arn=topic_arn)
        except ClientError:
            _LOG.warn('The final access policy size for lambda {} is reached. '
                      'The limit is 20480 bytes. '
                      'Invocation permission was not added.'.format(
                lambda_arn))

    def create_sns_subscription_for_lambda(self, lambda_arn, topic_name,
                                           region):
        """ Create subscription for lambda on SNS topic in specified
        region/regions.

        :type lambda_arn: str
        :type topic_name: str
        :type region: str
        """
        if region:
            if isinstance(region, str):
                if region == 'all':
                    for each in ALL_REGIONS:
                        self._subscribe_lambda_to_sns_topic(lambda_arn,
                                                            topic_name,
                                                            each)
                else:
                    if check_region_available(region, ALL_REGIONS):
                        self._subscribe_lambda_to_sns_topic(lambda_arn,
                                                            topic_name,
                                                            region)
            elif isinstance(region, list):
                for each in region:
                    if check_region_available(each, ALL_REGIONS):
                        self._subscribe_lambda_to_sns_topic(lambda_arn,
                                                            topic_name,
                                                            each)
            else:
                raise AssertionError('Invalid value for SNS region: %s.',
                                     region)
        else:
            self._subscribe_lambda_to_sns_topic(lambda_arn, topic_name,
                                                self.region)

    def _create_cloud_watch_trigger_from_meta(self, topic_name, trigger_meta,
                                              region):
        required_parameters = ['target_rule']
        validate_params(topic_name, trigger_meta, required_parameters)
        rule_name = trigger_meta['target_rule']

        topic_arn = self.connection_provider.sns(region).get_topic_arn(
            topic_name)
        self.connection.cw_events(region).add_rule_target(rule_name, topic_arn)
        self.connection_provider.sns(region).allow_service_invoke(
            topic_arn, 'events.amazonaws.com')
        _LOG.info('SNS topic %s subscribed to cloudwatch rule %s', topic_name,
                  rule_name)

    CREATE_TRIGGER = {
        'cloudwatch_rule_trigger': _create_cloud_watch_trigger_from_meta
    }

    def remove_sns_topics(self, args):
        self.create_pool(self._remove_sns_topic, args)

    @unpack_kwargs
    def _remove_sns_topic(self, arn, config):
        region = arn.split(':')[3]
        topic_name = config['resource_name']
        try:
            self.connection_provider.sns(region).remove_topic_by_arn(arn)
            _LOG.info('SNS topic %s was removed.', topic_name)
        except ClientError as e:
            exception_type = e.response['Error']['Code']
            if exception_type == 'ResourceNotFoundException':
                _LOG.warn('SNS topic %s is not found', topic_name)
            else:
                raise e

    @unpack_kwargs
    def _create_platform_application_from_meta(self, name, meta, region):
        required_parameters = ['platform', 'attributes']
        validate_params(name, meta, required_parameters)
        arn = self.connection_provider.sns(region).get_platform_application(
            name)
        if arn:
            _LOG.warn(
                '{0} SNS platform application exists in region {1}.'.format(
                    name, region))
            return self.describe_sns_application(name, meta, region, arn)
        platform = meta['platform']
        atrbts = meta['attributes']
        try:
            arn = self.connection_provider.sns(
                region).create_platform_application(
                name=name,
                platform=platform,
                attributes=atrbts)
        except ClientError as e:
            exception_type = e.response['Error']['Code']
            if exception_type == 'InvalidParameterException':
                _LOG.warn('SNS application %s is already exist.', name)
            else:
                raise e
        _LOG.info('SNS platform application %s in region %s has been created.',
                  name, region)
        return self.describe_sns_application(name, meta, region, arn)

    def remove_sns_application(self, args):
        self.create_pool(self._remove_sns_application, args)

    @unpack_kwargs
    def _remove_sns_application(self, arn, config):
        region = arn.split(':')[3]
        application_name = config['resource_name']
        try:
            self.connection_provider.sns(region).remove_application_by_arn(arn)
            _LOG.info('SNS application %s was removed.', application_name)
        except ClientError as e:
            exception_type = e.response['Error']['Code']
            if exception_type == 'ResourceNotFoundException':
                _LOG.warn('SNS application %s is not found', application_name)
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.kms_connection')


@apply_methods_decorator(retry)
class KMSConnection(object):
    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('kms', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new KMS connection.')

    def list_keys(self):
        keys = []
        response = self.client.list_keys()
        if 'Keys' in response:
            keys.extend(response.get('Keys'))
        while response.get('Truncated'):
            response = self.client.list_keys(Marker=response.get('NextMarker'))
            if 'Keys' in response:
                keys.extend(response.get('Keys'))
        return keys

    def describe_key(self, key_id):
        response = self.client.describe_key(KeyId=key_id)
        if response:
            return response.get('KeyMetadata')

    def create_key(self, description='', key_usage='ENCRYPT_DECRYPT',
                   origin='AWS_KMS', policy=None):
        arguments = {
            'Description': description,
            'KeyUsage': key_usage,
            'Origin': origin
        }
        if policy:
            arguments['Policy'] = policy

        response = self.client.create_key(**arguments)

        if response and response.get('KeyMetadata'):
            return response.get('KeyMetadata').get('KeyId')

    def schedule_key_deletion(self, key, days=7):
        self.client.schedule_key_deletion(KeyId=key, PendingWindowInDays=days)

    def list_aliases(self):
        aliases = []
        response = self.client.list_aliases()
        if 'Aliases' in response:
            aliases.extend(response.get('Aliases'))
        while response.get('Truncated'):
            response = self.client.list_aliases(
                Marker=response.get('NextMarker'))
            if 'Aliases' in response:
                aliases.extend(response.get('Aliases'))
        return aliases

    def create_alias(self, alias, key):
        self.client.create_alias(AliasName=alias, TargetKeyId=key)

    def update_alias(self, alias, key):
        self.client.update_alias(AliasName=alias, TargetKeyId=key)

    def delete_alias(self, alias):
        self.client.delete_alias(AliasName=alias)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.commons.log_helper import get_logger
from syndicate.core import ClientError
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj, chunks

_LOG = get_logger('syndicate.core.resources.s3_resource')


class S3Resource(BaseResource):

    def __init__(self, s3_conn) -> None:
        self.s3_conn = s3_conn

    def create_s3_bucket(self, args):
        return self.create_pool(self._create_s3_bucket_from_meta, args)

    def describe_bucket(self, name, meta):
        arn = 'arn:aws:s3:::{0}'.format(name)
        acl_response = self.s3_conn.get_bucket_acl(name)
        location_response = self.s3_conn.get_bucket_location(name)
        bucket_policy = self.s3_conn.get_bucket_policy(name)
        response = {
            'bucket_acl': acl_response,
            'location': location_response,
        }
        if bucket_policy:
            response['policy'] = bucket_policy
        return {
            arn: build_description_obj(response, name, meta)
        }

    @unpack_kwargs
    def _create_s3_bucket_from_meta(self, name, meta):
        if self.s3_conn.is_bucket_exists(name):
            _LOG.warn('{0} bucket exists.'.format(name))
            return self.describe_bucket(name, meta)
        self.s3_conn.create_bucket(name, meta.get('acl'), meta.get('location'))
        policy = meta.get('policy')
        if policy:
            self.s3_conn.add_bucket_policy(name, policy)
            _LOG.debug('Policy on {0} S3 bucket is set up.'.format(name))
        _LOG.info('Created S3 bucket {0}.'.format(name))
        rules = meta.get('LifecycleConfiguration')
        if rules:
            self.s3_conn.add_bucket_rule(name, rules)
            _LOG.debug('Rules on {0} S3 bucket are set up.'.format(name))
        cors_configuration = meta.get('cors')
        if cors_configuration:
            self.s3_conn.put_cors(bucket_name=name, rules=cors_configuration)
        return self.describe_bucket(name, meta)

    def _delete_objects(self, bucket_name, keys):
        response = self.s3_conn.delete_objects(bucket_name, keys)
        errors = response.get('Errors')
        if errors:
            error_keys = [{
                'Key': i['Key'],
                'VersionId': i['VersionId']
            } for i in errors]
            return error_keys
        else:
            return []

    def remove_buckets(self, args):
        self.create_pool(self._remove_bucket, args)

    @unpack_kwargs
    def _remove_bucket(self, arn, config):
        bucket_name = config['resource_name']
        try:
            errors = []
            keys = self.s3_conn.list_object_versions(bucket_name)
            if keys:
                for s3_keys in chunks(keys, 1000):
                    errors.extend(self._delete_objects(bucket_name, s3_keys))

            markers = self.s3_conn.list_object_markers(bucket_name)
            if markers:
                for s3_markers in chunks(markers, 1000):
                    errors.extend(
                        self._delete_objects(bucket_name, s3_markers))

            if errors:
                raise AssertionError('Error occurred while deleting S3 objects'
                                     ' from {0} bucket. Not deleted keys: '
                                     '{1}'.format(bucket_name, str(errors)))
            else:
                self.s3_conn.delete_bucket(bucket_name)
                _LOG.info('S3 bucket {0} was removed.'.format(bucket_name))
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchBucket':
                _LOG.warn('S3 bucket {0} is not found'.format(bucket_name))
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os

from syndicate.commons.log_helper import get_logger

_LOG = get_logger('syndicate.core.generators')

FILE_LAMBDA_HANDLER_PYTHON = '/handler.py'
FILE_LAMBDA_HANDLER_NODEJS = '/index.js'


def _touch(path):
    try:
        with open(path, 'a'):
            os.utime(path, None)
    except OSError:
        raise RuntimeError('Can not create new file by path {}. Syndicate '
                           'does not have enough permissions!'.format(path))


def _mkdir(path, exist_ok=False, fault_message=None):
    try:
        os.makedirs(path, exist_ok=exist_ok)
        return True
    except FileExistsError as e:
        if fault_message:
            answer = input(fault_message)
            return _re_survey(answer, path)
        else:
            _LOG.error(e)
    except OSError:
        raise RuntimeError('Can not create new folder by path {}. Syndicate '
                           'does not have enough permissions!'.format(path))


def _re_survey(answer, project_path):
    while answer not in ('y', 'n'):
        answer = input('Please enter [y/n] value: ')

        if answer == 'y':
            os.makedirs(project_path, exist_ok=True)
            return True
        elif answer == 'n':
            return False
    return True


def _write_content_to_file(file, content):
    with open(file, 'w') as f:
        f.write(content)


def _read_content_from_file(file):
    with open(file, 'r') as f:
        return f.read()


def _alias_variable(alias_name):
    return '${' + alias_name + '}'



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import collections
import concurrent.futures
import datetime
import json
import os
import subprocess
import sys
from functools import wraps
from threading import Thread
from time import time

import click
from click import BadParameter
from tqdm import tqdm

from syndicate.commons.log_helper import get_logger
from syndicate.core.conf.processor import path_resolver
from syndicate.core.constants import (ARTIFACTS_FOLDER, BUILD_META_FILE_NAME,
                                      DEFAULT_SEP)

_LOG = get_logger('syndicate.core.helper')

CONF_PATH = os.environ.get('SDCT_CONF')


def unpack_kwargs(handler_func):
    """ Decorator for unpack kwargs.

    :type handler_func: func
    :param handler_func: function which will be decorated
    """

    @wraps(handler_func)
    def wrapper(*kwargs):
        """ Wrapper func."""
        parameters = {}
        for i in kwargs:
            if type(i) is dict:
                parameters = i
        return handler_func(**parameters)

    return wrapper


def exit_on_exception(handler_func):
    """ Decorator to catch all exceptions and fail stage execution

    :type handler_func: func
    :param handler_func: function which will be decorated
    """

    @wraps(handler_func)
    def wrapper(*args, **kwargs):
        """ Wrapper func."""
        try:
            return handler_func(*args, **kwargs)
        except Exception as e:
            _LOG.exception("Error occurred: %s", str(e))
            sys.exit(1)

    return wrapper


def prettify_json(obj):
    return json.dumps(obj, indent=4)


def cli_command(handler_func):
    @wraps(handler_func)
    def wrapper(*args, **kwargs):
        status_code = handler_func(*args, **kwargs)
        if status_code != 0:
            _LOG.error('Execution is failed')
            sys.exit(1)

    return wrapper


@cli_command
def execute_command_by_path(command, path):
    return subprocess.call(command, shell=True, cwd=path)


@cli_command
def execute_command(command):
    return subprocess.call(command, shell=True)


def build_path(*paths):
    return DEFAULT_SEP.join(paths)


def _find_alias_and_replace(some_string):
    """ Find placeholder for alias in string. If found - replace with alias
    value.

    :type some_string: str
    """
    from syndicate.core import CONFIG
    first_index = some_string.index('${')
    second_index = some_string.index('}')
    alias_name = some_string[first_index + 2:second_index]
    res_alias = CONFIG.resolve_alias(alias_name)
    if not res_alias:
        raise AssertionError('Can not found alias for {0}'.format(alias_name))
    result = (
            some_string[:first_index] + res_alias + some_string[
                                                    second_index + 1:])
    return result


def resolve_aliases_for_string(string_value):
    """ Look for aliases in string.

    :type string_value: str
    """
    input_string = string_value
    try:
        if '${' in string_value:
            if string_value.count('${') == string_value.count('}'):
                while True:
                    input_string = _find_alias_and_replace(input_string)
            else:
                raise AssertionError('Broken alias in value: {0}.'.format(
                    string_value))
        return input_string
    except ValueError:
        return input_string


def check_required_param(ctx, param, value):
    if not value:
        raise BadParameter('Parameter is required')
    return value


def resolve_path_callback(ctx, param, value):
    if not value:
        raise BadParameter('Parameter is required')
    return path_resolver(value)


def create_bundle_callback(ctx, param, value):
    from syndicate.core import CONFIG
    bundle_path = os.path.join(CONFIG.project_path, ARTIFACTS_FOLDER, value)
    if not os.path.exists(bundle_path):
        os.makedirs(bundle_path)
    return value


def verify_bundle_callback(ctx, param, value):
    from syndicate.core import CONFIG
    bundle_path = os.path.join(CONFIG.project_path, ARTIFACTS_FOLDER, value)
    if not os.path.exists(bundle_path):
        raise AssertionError("Bundle name does not exist. Please, invoke "
                             "'build_artifacts' command to create a bundle.")
    return value


def verify_meta_bundle_callback(ctx, param, value):
    bundle_path = build_path(CONF_PATH, ARTIFACTS_FOLDER, value)
    build_meta_path = os.path.join(bundle_path, BUILD_META_FILE_NAME)
    if not os.path.exists(build_meta_path):
        raise AssertionError(
            "Bundle name is incorrect. {0} does not exist. Please, invoke "
            "'package_meta' command to create a file.".format(
                BUILD_META_FILE_NAME))
    return value


def write_content_to_file(file_path, file_name, obj):
    file_name = os.path.join(file_path, file_name)
    if os.path.exists(file_name):
        _LOG.warn('{0} already exists'.format(file_name))
    else:
        with open(file_name, 'w') as meta_file:
            json.dump(obj, meta_file)
        _LOG.info('{0} file was created.'.format(meta_file.name))


def timeit(handler_func):
    @wraps(handler_func)
    def timed(*args, **kwargs):
        ts = time()
        result = handler_func(*args, **kwargs)
        te = time()
        _LOG.info('Stage %s, elapsed time: %s', handler_func.__name__,
                  str(datetime.timedelta(seconds=te - ts)))
        return result

    return timed


def execute_parallel_tasks(*fns):
    threads = []
    for fn in fns:
        t = Thread(target=fn)
        t.start()
        threads.append(t)
    for t in threads:
        t.join()


def handle_futures_progress_bar(futures):
    kwargs = {
        'total': len(futures),
        'unit': 'nap',
        'leave': True
    }
    for _ in tqdm(concurrent.futures.as_completed(futures), **kwargs):
        pass


class OrderedGroup(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        super(OrderedGroup, self).__init__(name, commands, **attrs)
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx):
        return self.commands



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import traceback
from functools import wraps
from time import sleep

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger

_LOG = get_logger('syndicate.connection.helper')


def apply_methods_decorator(decorator):
    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate


def retry(handler_func):
    """ Decorator for retry on specified exceptions.

    :type handler_func: func
    :param handler_func: function which will be decorated
    """

    # TODO check if it is possible to use it like this: @retry(on='blahblah')
    @wraps(handler_func)
    def wrapper(*args, **kwargs):
        """ Wrapper func."""
        retry_exceptions = [
            'ThrottlingException',
            'LimitExceededException',
            'ProvisionedThroughputExceededException',
            'TooManyRequestsException',
            'ConflictException',
            'An error occurred (InvalidParameterValueException) when calling'
            'CreateEventSourceMapping operation',
            'The role defined for the function cannot be assumed by Lambda',
            'An error occurred (ResourceConflictException) when calling'
            ' the AddPermission operation: The statement id',
            'NoSuchUpload',
            'Throttling',
            'Please add Lambda as a Trusted Entity',
            'UpdateFunctionConfiguration',
            'PutScalingPolicy',
            'RegisterScalableTarget',
            'TopicArn can not be None',
            'DeleteRole',
            'Max attempts exceeded',
            'UpdateGatewayResponse'
        ]
        for each in range(1, 20, 3):
            try:
                return handler_func(*args, **kwargs)
            except ClientError as e:
                retry_flag = False
                for exc in retry_exceptions:
                    if exc in str(e):
                        _LOG.debug('Retry on {0}. Error: {1}'.format(
                            handler_func.__name__, str(e)))
                        _LOG.debug('Parameters: {0}, {1}'.format(str(args),
                                                                 str(kwargs)))
                        # set to debug, we need it only in the logs file
                        _LOG.debug(
                            'Traceback:\n {0}'.format(traceback.format_exc()))
                        retry_flag = True
                if not retry_flag:
                    raise e
                sleep(each)

    return wrapper



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.conf.processor import ALL_REGIONS
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             create_args_for_multi_region,
                                             validate_params)

_LOG = get_logger('syndicate.core.resources.cloud_watch_resource')
ARN_KEY = 'Arn'


def _create_ec2_rule(rule_name, rule_meta, cw_conn):
    cw_conn.create_ec2_rule(name=rule_name,
                            instances=rule_meta.get('instance_ids'),
                            instance_states=rule_meta.get(
                                'instance_states'))


def _create_schedule_rule(rule_name, rule_meta, cw_conn):
    cw_conn.create_schedule_rule(name=rule_name,
                                 expression=rule_meta['expression'])


def _create_api_call_rule(rule_name, rule_meta, cw_conn):
    cw_conn.create_api_call_rule(name=rule_name,
                                 aws_service=rule_meta.get('aws_service'),
                                 operations=rule_meta.get('operations'),
                                 custom_pattern=rule_meta.get(
                                     'custom_pattern'))


RULE_TYPES = {
    'schedule': _create_schedule_rule,
    'ec2': _create_ec2_rule,
    'api_call': _create_api_call_rule
}


class CloudWatchResource(BaseResource):

    def __init__(self, cw_events_conn_builder, account_id) -> None:
        self.cw_events_conn = cw_events_conn_builder()
        self._cw_events_conn_builder = cw_events_conn_builder
        self.account_id = account_id

    def describe_rule(self, name, meta, region, response=None):
        if not response:
            response = self._cw_events_conn_builder(region).get_rule(name)
        arn = response[ARN_KEY]
        del response[ARN_KEY]
        return {arn: build_description_obj(response, name, meta)}

    def describe_rule_from_meta(self, name, meta):
        new_region_args = create_args_for_multi_region(
            [
                {'name': name,
                 'meta': meta}
            ], ALL_REGIONS)
        responses = []
        for arg in new_region_args:
            rule = self._cw_events_conn_builder(arg['region']).get_rule(name)
            if rule:
                responses.append(rule)

        description = []
        for rule in responses:
            arn = rule[ARN_KEY]
            del rule[ARN_KEY]
            description.append({arn: build_description_obj(rule, name, meta)})
        return description

    def create_cloud_watch_rule(self, args):
        """ Create CloudWatch rule from meta in region/regions.

        :type args: list
        """
        new_region_args = create_args_for_multi_region(args, ALL_REGIONS)
        return self.create_pool(self._create_cloud_watch_rule_from_meta,
                                new_region_args)

    @unpack_kwargs
    def _create_cloud_watch_rule_from_meta(self, name, meta, region):
        # validation depends on rule type
        required_parameters = ['rule_type']
        rule_type = meta.get('rule_type')
        if rule_type:
            if rule_type == 'schedule':
                required_parameters.append('expression')
        validate_params(name, meta, required_parameters)

        event_buses = meta.get('event_bus_accounts')
        response = self._cw_events_conn_builder(region).get_rule(name)
        if response:
            _LOG.warn('%s rule exists in %s.', name, region)
            return self.describe_rule(name=name, meta=meta, region=region,
                                      response=response)
        try:
            func = RULE_TYPES[rule_type]
            func(name, meta, self._cw_events_conn_builder(region))
            if event_buses:
                time.sleep(5)
                self._attach_tenant_rule_targets(name, region, event_buses)
            _LOG.info('Created cloud watch rule %s in %s.', name, region)
            response = self._cw_events_conn_builder(region).get_rule(name)
            time.sleep(5)
            return self.describe_rule(name=name, meta=meta, region=region,
                                      response=response)
        except KeyError:
            raise AssertionError(
                'Invalid rule type: {0} for resource {1}. '
                'Please, change rule type with existing: '
                'schedule|ec2|api_call.'.format(rule_type, name))

    def _attach_tenant_rule_targets(self, rule_name, region, event_buses):
        for event_bus in event_buses:
            target_arn = 'arn:aws:events:{0}:{1}:event-bus/default'.format(
                region,
                event_bus)
            existing_targets = self.cw_events_conn(
                region).list_targets_by_rule(
                rule_name=rule_name)
            for target in existing_targets:
                if target[ARN_KEY] == target_arn:
                    _LOG.debug('Target to event bus %s is already attached',
                               target_arn)
                    return
            self._cw_events_conn_builder(region).add_rule_target(
                rule_name=rule_name,
                target_arn=target_arn)

    def _handle_deactivation_for_cw_resources(self, cw_conn, region,
                                              rule_name):
        targets = cw_conn.list_targets_by_rule(rule_name)
        home_eb_arn = f'arn:aws:events:' \
                      f'{region}:{self.account_id}:event-bus/default'
        _LOG.debug('Home account event bus arn: %s', home_eb_arn)
        for target in targets:
            resource_arn = target[ARN_KEY]
            if resource_arn == home_eb_arn:
                cw_conn.remove_targets(rule_name, [target['Id']])
                _LOG.debug('Target %s removed', resource_arn)
        targets = cw_conn.list_targets_by_rule(rule_name)
        if targets:
            _LOG.debug('Will not remove rule, targets attached')
        else:
            _LOG.debug('Going to remove rule %s', rule_name)
            cw_conn.remove_rule(rule_name)
            _LOG.debug('Rule %s removed', rule_name)

    def remove_cloud_watch_rules(self, args):
        self.create_pool(self._remove_cloud_watch_rule, args)

    @unpack_kwargs
    def _remove_cloud_watch_rule(self, arn, config):
        region = arn.split(':')[3]
        resource_name = config['resource_name']
        try:
            self._cw_events_conn_builder(region).remove_rule(resource_name)
            _LOG.info('Rule %s was removed', resource_name)
        except ClientError as e:
            exception_type = e.response['Error']['Code']
            if exception_type == 'ResourceNotFoundException':
                _LOG.warn('Rule %s is not found', resource_name)
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.commons.log_helper import get_logger
from syndicate.core.conf.processor import GLOBAL_AWS_SERVICES

_LOG = get_logger('syndicate.core.resources.helper')


def validate_params(name, meta, required_params):
    existing_parameters = list(meta.keys())
    parameters_string = ', '.join(required_params)
    existing_parameters_string = ', '.join(existing_parameters)
    for each in required_params:
        if each not in existing_parameters:
            raise AssertionError(
                'All required parameters must be specified! Resource: {0}'
                ' Required parameters: {1}. Given parameters: {2}'.format(
                    name, parameters_string, existing_parameters_string))


def check_region_available(region_name, available_regions, res_meta=None):
    if region_name in available_regions:
        return True
    if res_meta:
        res_type = res_meta['resource_type']
        raise AssertionError(
            "Region {0} isn't available for resource {1}.".format(region_name,
                                                                  res_type))
    else:
        raise AssertionError("Region {0} isn't available.".format(region_name))


def create_args_for_multi_region(args, available_regions):
    from syndicate.core import CONFIG
    new_region_args = []
    for arg_set in args:
        name = arg_set['name']
        meta = arg_set['meta']
        region = meta.get('region')
        if region is None:
            item = arg_set.copy()
            item['region'] = CONFIG.region
            new_region_args.append(item)
        elif isinstance(region, str):
            if region == 'all':
                for each in available_regions:
                    item = arg_set.copy()
                    item['region'] = each
                    new_region_args.append(item)
            else:
                if check_region_available(region, available_regions, meta):
                    item = arg_set.copy()
                    item['region'] = region
                    new_region_args.append(item)
        elif isinstance(region, list):
            for each in region:
                if check_region_available(each, available_regions, meta):
                    item = arg_set.copy()
                    item['region'] = each
                    new_region_args.append(item)
        else:
            raise AssertionError(
                'Invalid value region: {0}. Resource: {1}.'.format(region,
                                                                   name))
    return new_region_args


def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i:i + n]


def resolve_dynamic_identifier(name, value, dict):
    return json.loads(json.dumps(dict).replace(name, value))


def build_description_obj(response, name, meta):
    resource_type = meta['resource_type']
    obj = {
        'resource_name': name,
        'resource_meta': meta,
        'description': response
    }
    if resource_type not in GLOBAL_AWS_SERVICES:
        from syndicate.core import CONFIG
        obj['resource_meta']['region'] = meta.get('region', CONFIG.region)
    return obj


def assert_required_params(required_params_names, all_params):
    """
    Raises error if there is at least one missing parameter.
    :param required_params_names:
    :param all_params:
    :return:
    """
    missing = [param for param in required_params_names if
               param not in all_params.keys()]
    if missing:
        raise AssertionError(f'Missing required parameters: {missing}')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.sqs_connection')
FIFO_REGIONS = ['us-east-1', 'us-east-2', 'us-west-2', 'eu-west-1']


@apply_methods_decorator(retry)
class SqsConnection(object):
    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.client = client('sqs', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new SQS connection.')

    def create_queue(self, queue_name, delay_seconds=None,
                     maximum_message_size=None, message_retention_period=None,
                     policy=None, receive_message_wait_time_seconds=None,
                     redrive_policy=None, visibility_timeout=None,
                     kms_master_key_id=None,
                     kms_data_key_reuse_period_seconds=None, fifo_queue=False,
                     content_based_deduplication=None):
        attributes = dict()
        if fifo_queue:
            if self.region not in FIFO_REGIONS:
                raise AssertionError(
                    'FIFO queue is not supported for {0} region.'
                    ' Available in : {1}'.format(self.region, FIFO_REGIONS))
            attributes['FifoQueue'] = str(fifo_queue)
        params = dict(QueueName=queue_name)
        if delay_seconds:
            if delay_seconds < 0 or delay_seconds > 900:
                raise AssertionError(
                    'Delay seconds for queue must be between 0 and 900 seconds')
            attributes['DelaySeconds'] = str(delay_seconds)
        if maximum_message_size:
            if maximum_message_size < 1024 or maximum_message_size > 262144:
                raise AssertionError(
                    'Maximim message size must be between 1024 and 262144 bytes')
            attributes['MaximumMessageSize'] = str(maximum_message_size)
        if message_retention_period:
            if message_retention_period < 60 or message_retention_period > 1209600:
                raise AssertionError(
                    'Message retention size must be between 60 and 1209600 seconds')
            attributes['MessageRetentionPeriod'] = str(
                message_retention_period)
        if policy:
            attributes['Policy'] = policy
        if receive_message_wait_time_seconds:
            if receive_message_wait_time_seconds < 0 or receive_message_wait_time_seconds > 20:
                raise AssertionError(
                    'Receive message wait time must be between 0 and 20 seconds')
            attributes[
                'ReceiveMessageWaitTimeSeconds'] = str(
                receive_message_wait_time_seconds)
        if redrive_policy:
            attributes['RedrivePolicy'] = redrive_policy
        if visibility_timeout:
            if visibility_timeout < 0 or visibility_timeout > 43200:
                raise AssertionError(
                    'Visibility timeout must be between 0 and 43200 seconds')
            attributes['VisibilityTimeout'] = str(visibility_timeout)
        if kms_master_key_id:
            attributes['KmsMasterKeyId'] = kms_master_key_id
        if kms_data_key_reuse_period_seconds:
            if kms_data_key_reuse_period_seconds < 60 or kms_data_key_reuse_period_seconds > 86400:
                raise AssertionError(
                    'KMS key reuse period must be between 60 and 86400 seconds')
            attributes[
                'KmsDataKeyReusePeriodSeconds'] = str(
                kms_data_key_reuse_period_seconds)
        if content_based_deduplication:
            attributes[
                'ContentBasedDeduplication'] = str(content_based_deduplication)
        params['Attributes'] = attributes
        return self.client.create_queue(**params)

    def delete_queue(self, queue_url):
        self.client.delete_queue(QueueUrl=queue_url)

    def list_queues(self, url_prefix):
        response = self.client.list_queues(QueueNamePrefix=url_prefix)
        return response.get('QueueUrls', list())

    def get_queue_url(self, queue_name, account_id):
        try:
            response = self.client.get_queue_url(QueueName=queue_name,
                                                 QueueOwnerAWSAccountId=account_id)
            return response['QueueUrl']
        except ClientError as e:
            if 'AWS.SimpleQueueService.NonExistentQueue' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_queue_attributes(self, queue_url, attribute_names=None):
        if attribute_names is None:
            attribute_names = ['All']
        try:
            return self.client.get_queue_attributes(QueueUrl=queue_url,
                                                    AttributeNames=attribute_names)
        except ClientError as e:
            if 'QueueDoesNotExistException' in str(e):
                pass  # valid exception
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import sys

import yaml

from syndicate.commons.log_helper import get_logger
from syndicate.core.conf.processor import (PROJECT_PATH_CFG,
                                           LEGACY_CONFIG_FILE_NAME,
                                           CONFIG_FILE_NAME,
                                           ALIASES_FILE_NAME,
                                           ACCOUNT_ID_CFG, REGION_CFG,
                                           DEPLOY_TARGET_BUCKET_CFG,
                                           AWS_ACCESS_KEY_ID_CFG,
                                           AWS_SECRET_ACCESS_KEY_CFG,
                                           PROJECTS_MAPPING_CFG,
                                           RESOURCES_PREFIX_CFG,
                                           RESOURCES_SUFFIX_CFG)
from syndicate.core.conf.validator import (PYTHON_BUILD_TOOL_NAME,
                                           MVN_BUILD_TOOL_NAME,
                                           NODE_BUILD_TOOL_NAME,
                                           LAMBDAS_ALIASES_NAME_CFG)

_LOG = get_logger('config_generator')


def generate_configuration_files(config_path, region, account_id,
                                 access_key, secret_key,
                                 bundle_bucket_name, python_build_mapping,
                                 java_build_mapping,
                                 nodejs_build_mapping, prefix, suffix,
                                 project_path=None):
    if not config_path:
        _LOG.warn(f'The {config_path} property is not specified. '
                  f'The working directory is used')
        config_path = os.getcwd()

    if not os.path.exists(config_path):
        try:
            os.mkdir(config_path)
        except OSError:
            _LOG.exception(
                f'Error has occurred while creating '
                f'directory for configs {config_path}')
            sys.exit(1)

    if not project_path:
        _LOG.warn(f'The {PROJECT_PATH_CFG} property is not specified. '
                  f'The working directory will be used as a project path. '
                  f'To change the path, edit the {LEGACY_CONFIG_FILE_NAME} '
                  f'by path {config_path}')
        project_path = os.getcwd()
    else:
        if not os.path.exists(project_path):
            raise AssertionError(
                f'Provided project path {project_path} does not exists')

    build_project_mapping = {}
    if python_build_mapping:
        build_project_mapping[PYTHON_BUILD_TOOL_NAME] = \
            list(python_build_mapping)
    if java_build_mapping:
        build_project_mapping[MVN_BUILD_TOOL_NAME] = list(java_build_mapping)
    if nodejs_build_mapping:
        build_project_mapping[NODE_BUILD_TOOL_NAME] = \
            list(nodejs_build_mapping)

    config_content = {
        ACCOUNT_ID_CFG: account_id,
        REGION_CFG: region,
        DEPLOY_TARGET_BUCKET_CFG: bundle_bucket_name,
        AWS_ACCESS_KEY_ID_CFG: access_key,
        AWS_SECRET_ACCESS_KEY_CFG: secret_key,
        PROJECT_PATH_CFG: project_path,
        PROJECTS_MAPPING_CFG: build_project_mapping,
        RESOURCES_PREFIX_CFG: prefix,
        RESOURCES_SUFFIX_CFG: suffix
    }

    config_file_path = f'{config_path}/{CONFIG_FILE_NAME}'
    with open(config_file_path, 'w') as config_file:
        yaml.dump(config_content, config_file)

    aliases_content = {
        ACCOUNT_ID_CFG: account_id,
        REGION_CFG: region,
        LAMBDAS_ALIASES_NAME_CFG: 'prod'
    }
    aliases_file_path = f'{config_path}/{ALIASES_FILE_NAME}'
    with open(aliases_file_path, 'w') as aliases_file:
        yaml.dump(aliases_content, aliases_file)

    _LOG.info(
        'Syndicate initialization has been completed. '
        f'Set SDCT_CONF:\nexport SDCT_CONF={config_path}')


def generate_build_project_mapping(mapping_item, build_type):
    type_mapping = ''
    for item in mapping_item:
        type_mapping += f'{build_type}:{item}'
    return type_mapping



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time

from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             validate_params)

_LOG = get_logger('core.resources.step_function_resource')


class StepFunctionResource(BaseResource):

    def __init__(self, sf_conn, iam_conn, cw_events_conn, lambda_conn,
                 account_id, region) -> None:
        self.sf_conn = sf_conn
        self.iam_conn = iam_conn
        self.cw_events_conn = cw_events_conn
        self.lambda_conn = lambda_conn
        self.account_id = account_id
        self.region = region

    def create_state_machine(self, args):
        return self.create_pool(self._create_state_machine_from_meta, args)

    def create_activities(self, args):
        return self.create_pool(self._create_activity_from_meta, args)

    def remove_state_machines(self, args):
        self.create_pool(self._remove_state_machine, args)
        if args:
            time.sleep(60)

    @unpack_kwargs
    def _remove_state_machine(self, arn, config):
        sm_name = config['resource_name']
        try:
            executions = self.sf_conn.list_executions_by_status(arn, 'RUNNING')
            if executions:
                _LOG.debug('Found {0} running executions '
                           'for {1}'.format(len(executions), sm_name))
                for execution in executions:
                    self.sf_conn.stop_execution(execution['executionArn'])
                _LOG.debug('Executions stop initiated')
            self.sf_conn.delete_state_machine(arn)
            _LOG.info('State machine %s was removed', sm_name)
        except ClientError as e:
            exception_type = e.response['Error']['Code']
            if exception_type == 'StateMachineDoesNotExist':
                _LOG.warn('State machine %s is not found', sm_name)
            else:
                raise e

    def remove_activities(self, args):
        self.create_pool(self._remove_activity, args)

    @unpack_kwargs
    def _remove_activity(self, arn, config):
        activity_name = config['resource_name']
        try:
            self.sf_conn.delete_activity(arn)
            _LOG.info('State activity %s was removed', activity_name)
        except ClientError as e:
            exception_type = e.response['Error']['Code']
            if exception_type == 'ResourceNotFoundException':
                _LOG.warn('State activity %s is not found', activity_name)
            else:
                raise e

    def __remove_key_from_dict(self, obj, name):
        try:
            del obj[name]
        except KeyError:
            pass

    @unpack_kwargs
    def _create_state_machine_from_meta(self, name, meta):
        arn = self._build_sm_arn(name, self.region)
        response = self.sf_conn.describe_state_machine(arn)
        if response:
            _LOG.warn('State machine %s exists', name)
            return {
                arn: build_description_obj(response, name, meta)
            }

        iam_role = meta['iam_role']
        role_arn = self.iam_conn.check_if_role_exists(iam_role)
        if not role_arn:
            raise AssertionError(
                'IAM role {0} does not exist.'.format(iam_role))

        # check resource exists and get arn
        definition = meta['definition']
        definition_copy = definition.copy()
        for key in definition['States']:
            definition_meta = definition['States'][key]
            if definition_meta.get('Lambda'):
                lambda_name = definition_meta['Lambda']
                # alias has a higher priority than version in arn resolving
                lambda_version = definition_meta.get('Lambda_version')
                lambda_alias = definition_meta.get('Lambda_alias')
                lambda_arn = self.resolve_lambda_arn_by_version_and_alias(
                    lambda_name,
                    lambda_version,
                    lambda_alias)
                self.__remove_key_from_dict(definition_copy['States'][key],
                                            'Lambda')
                self.__remove_key_from_dict(definition_copy['States'][key],
                                            'Lambda_version')
                self.__remove_key_from_dict(definition_copy['States'][key],
                                            'Lambda_alias')

                definition_copy['States'][key]['Resource'] = lambda_arn

            if definition_meta.get('Activity'):
                activity_name = definition_meta['Activity']
                activity_arn = 'arn:aws:states:{0}:{1}:activity:{2}'.format(
                    self.region, self.account_id, activity_name)
                activity_info = self.sf_conn.describe_activity(
                    arn=activity_arn)
                if not activity_info:
                    raise AssertionError('Activity does not exists: %s',
                                         activity_name)
                activity_arn = activity_info['activityArn']
                del definition_copy['States'][key]['Activity']
                definition_copy['States'][key]['Resource'] = activity_arn
        machine_info = self.sf_conn.create_state_machine(machine_name=name,
                                                         role_arn=role_arn,
                                                         definition=definition_copy)

        event_sources = meta.get('event_sources')
        if event_sources:
            for trigger_meta in event_sources:
                trigger_type = trigger_meta['resource_type']
                func = self.CREATE_TRIGGER[trigger_type]
                func(name, trigger_meta)
        _LOG.info('Created state machine %s.', machine_info['stateMachineArn'])
        return self.describe_step_function(name=name, meta=meta, arn=arn)

    def describe_step_function(self, name, meta, arn=None):
        if not arn:
            arn = self._build_sm_arn(name, self.region)
        response = self.sf_conn.describe_state_machine(arn)
        return {
            arn: build_description_obj(response, name, meta)
        }

    @staticmethod
    def _build_sm_arn(self, name, region):
        return f'arn:aws:states:{region}:{self.account_id}:stateMachine:{name}'

    def _create_cloud_watch_trigger_from_meta(self, name, trigger_meta):
        required_parameters = ['target_rule', 'input', 'iam_role']
        validate_params(name, trigger_meta, required_parameters)
        rule_name = trigger_meta['target_rule']
        input = trigger_meta['input']
        sf_role = trigger_meta['iam_role']

        sf_arn = self._build_sm_arn(name, self.region)
        sf_description = self.sf_conn.describe_state_machine(arn=sf_arn)
        if sf_description.get('status') == 'ACTIVE':
            sf_role_arn = self.iam_conn.check_if_role_exists(sf_role)
            if sf_role_arn:
                self.cw_events_conn.add_rule_sf_target(rule_name, sf_arn,
                                                       input,
                                                       sf_role_arn)
                _LOG.info('State machine %s subscribed to cloudwatch rule %s',
                          name, rule_name)

    CREATE_TRIGGER = {
        'cloudwatch_rule_trigger': _create_cloud_watch_trigger_from_meta
    }

    @unpack_kwargs
    def _create_activity_from_meta(self, name, meta):
        arn = self.build_activity_arn(name=name)
        response = self.sf_conn.describe_activity(arn)
        if response:
            _LOG.warn('Activity %s exists.', name)
            return {
                arn: build_description_obj(response, name, meta)
            }
        response = self.sf_conn.create_activity(name=name)
        _LOG.info('Activity %s is created.', name)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def describe_activity(self, name, meta):
        arn = self.build_activity_arn(name=name)
        response = self.sf_conn.describe_activity(arn=arn)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def build_activity_arn(self, name):
        arn = 'arn:aws:states:{0}:{1}:activity:{2}'.format(self.region,
                                                           self.account_id,
                                                           name)
        return arn



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os

from syndicate.commons.log_helper import get_logger
from syndicate.core.generators import (_touch,
                                       _mkdir, _write_content_to_file,
                                       FILE_LAMBDA_HANDLER_PYTHON,
                                       FILE_LAMBDA_HANDLER_NODEJS,
                                       _read_content_from_file)
from syndicate.core.generators.contents import (
    NODEJS_LAMBDA_HANDLER_TEMPLATE,
    PYTHON_LAMBDA_HANDLER_TEMPLATE,
    _generate_python_node_lambda_config,
    _generate_lambda_role_config, _generate_nodejs_node_lambda_config,
    CANCEL_MESSAGE, _generate_package_nodejs_lambda,
    _generate_package_lock_nodejs_lambda, JAVA_LAMBDA_HANDLER_CLASS,
    SRC_MAIN_JAVA, FILE_POM)
from syndicate.core.groups import (PROJECT_JAVA, PROJECT_NODEJS,
                                   PROJECT_PYTHON)

_LOG = get_logger('syndicate.core.generators.lambda_function')

SLASH_SYMBOL = '/'

FOLDER_LAMBDAS = '/lambdas'
FOLDER_COMMONS = '/commons'

FILE_README = '/README.md'
FILE_DEPLOYMENT_RESOURCES = '/deployment_resources.json'
FILE_INIT_PYTHON = '/__init__.py'
FILE_LAMBDA_CONFIG = '/lambda_config.json'
FILE_PACKAGE_LOCK = '/package-lock.json'
FILE_PACKAGE = '/package.json'

FILE_REQUIREMENTS = '/requirements.txt'
FILE_LOCAL_REQUIREMENTS = '/local_requirements.txt'

LAMBDA_ROLE_NAME_PATTERN = '{0}-role'  # 0 - lambda_name
POLICY_NAME_PATTERN = '{0}-policy'  # 0 - lambda_name

PYTHON_LAMBDA_FILES = [
    FILE_INIT_PYTHON, FILE_LOCAL_REQUIREMENTS,
    FILE_REQUIREMENTS,
    FILE_DEPLOYMENT_RESOURCES,  # content
    FILE_LAMBDA_CONFIG  # content
]  # + handler content

NODEJS_LAMBDA_FILES = [
    FILE_PACKAGE,
    FILE_PACKAGE_LOCK,
    FILE_LAMBDA_CONFIG,
    FILE_LAMBDA_HANDLER_NODEJS,
    FILE_DEPLOYMENT_RESOURCES
]


def generate_lambda_function(project_name, project_path, project_language,
                             lambda_names):
    full_project_path = os.path.join(project_path, project_name)

    if not os.path.exists(full_project_path):
        _LOG.info('Project "{}" you have provided does not exist'.format(
            full_project_path))
        return

    processor = LAMBDAS_PROCESSORS.get(project_language)
    if not processor:
        raise RuntimeError('Wrong project language {0}'.format(
            project_language))

    lambdas_path = full_project_path + FOLDER_LAMBDAS

    processor(project_name=project_name, project_path=project_path,
              lambda_names=lambda_names, lambdas_path=lambdas_path)
    _LOG.info(f'Lambda generating have been successfully performed.')


def _generate_python_lambdas(lambda_names, lambdas_path, project_name=None,
                             project_path=None):
    if not os.path.exists(lambdas_path):
        _mkdir(lambdas_path, exist_ok=True)
    for lambda_name in lambda_names:
        print(lambdas_path)
        lambda_folder = os.path.join(lambdas_path, lambda_name)

        answer = _mkdir(
            path=lambda_folder,
            fault_message=f'\nLambda {lambda_name} already exists.\nOverride the '
            'Lambda function? [y/n]: ')
        if not answer:
            _LOG.info(CANCEL_MESSAGE.format(lambda_name))
            continue

        PYTHON_LAMBDA_FILES.append(
            FILE_LAMBDA_HANDLER_PYTHON)  # add lambda handler
        for file in PYTHON_LAMBDA_FILES:
            _touch(lambda_folder + file)

        # fill handler.py
        _write_content_to_file(f'{lambda_folder}/{FILE_LAMBDA_HANDLER_PYTHON}',
                               PYTHON_LAMBDA_HANDLER_TEMPLATE)

        # fill deployment_resources.json
        pattern_format = LAMBDA_ROLE_NAME_PATTERN.format(lambda_name)
        role_def = _generate_lambda_role_config(
            pattern_format)
        _write_content_to_file(f'{lambda_folder}/{FILE_DEPLOYMENT_RESOURCES}',
                               role_def)

        # fill lambda_config.json
        lambda_def = _generate_python_node_lambda_config(
            lambda_name,
            f'{FOLDER_LAMBDAS}/{lambda_name}')
        _write_content_to_file(f'{lambda_folder}/{FILE_LAMBDA_CONFIG}',
                               lambda_def)
        _LOG.info(f'Lambda {lambda_name} created')


def _generate_java_lambdas(project_path, project_name, lambda_names,
                           lambdas_path):
    unified_lambda_name = _get_parts_split_by_chars(to_split=project_name,
                                                    chars=['-', '_'])
    java_package_name = unified_lambda_name.replace(' ', '')
    java_package_name = f'com.{java_package_name}'
    java_package_as_path = java_package_name.replace('.', '/')

    pom_file_path = f'{project_path}/{project_name}/{FILE_POM}'
    pom_xml_content = _read_content_from_file(pom_file_path)
    pom_xml_content = pom_xml_content.replace(
        '<!--packages to scan-->',
        f'<package>{java_package_name}</package>')
    _write_content_to_file(pom_file_path, pom_xml_content)

    full_package_path = f'{project_path}/{project_name}/{SRC_MAIN_JAVA}/' \
        f'{java_package_as_path}'
    for lambda_name in lambda_names:
        if not os.path.exists(full_package_path):
            _mkdir(full_package_path, exist_ok=True)

        lambda_class_name = unified_lambda_name.title()
        lambda_class_name = lambda_class_name.replace(' ', '')
        java_handler_content = JAVA_LAMBDA_HANDLER_CLASS.replace(
            '{java_package_name}',
            java_package_name
        )
        java_handler_content = java_handler_content.replace(
            '{lambda_name}',
            lambda_name
        )
        java_handler_content = java_handler_content.replace(
            '{lambda_class_name}',
            lambda_class_name
        )
        lambda_role_name = LAMBDA_ROLE_NAME_PATTERN.format(lambda_name)
        java_handler_content = java_handler_content.replace(
            '{lambda_role_name}',
            lambda_role_name
        )
        java_handler_file_name = f'{project_path}/{project_name}/' \
            f'{SRC_MAIN_JAVA}/{java_package_as_path}/{lambda_class_name}.java'
        _write_content_to_file(
            java_handler_file_name,
            java_handler_content
        )

        # add role to deployment_resource.json

        dep_res_path = f'{project_path}/{project_name}' \
            f'/{FILE_DEPLOYMENT_RESOURCES}'
        deployment_resources = json.loads(_read_content_from_file(
            dep_res_path
        ))
        deployment_resources.update(_generate_lambda_role_config(
            lambda_role_name, stringify=False))
        _write_content_to_file(dep_res_path,
                               json.dumps(deployment_resources, indent=2))

        _LOG.info(f'Lambda {lambda_name} created')


def _get_parts_split_by_chars(chars, to_split):
    result = to_split
    for char in chars:
        result = result.replace(char, ' ')
    return result


def _generate_nodejs_lambdas(lambda_names, lambdas_path, project_name=None,
                             project_path=None):
    if not os.path.exists(lambdas_path):
        _mkdir(lambdas_path, exist_ok=True)
    for lambda_name in lambda_names:

        lambda_folder = os.path.join(lambdas_path, lambda_name)

        answer = _mkdir(
            path=lambda_folder,
            fault_message=f'\nLambda {lambda_name} already exists.\n'
            f'Override the Lambda function? [y/n]: ')
        if not answer:
            _LOG.info(CANCEL_MESSAGE.format(lambda_name))
            continue

        for file in NODEJS_LAMBDA_FILES:
            _touch(lambda_folder + file)

        # fill index.js
        _write_content_to_file(f'{lambda_folder}/{FILE_LAMBDA_HANDLER_NODEJS}',
                               NODEJS_LAMBDA_HANDLER_TEMPLATE)

        # fill package.json
        package_def = _generate_package_nodejs_lambda(lambda_name)
        _write_content_to_file(f'{lambda_folder}/'f'{FILE_PACKAGE}',
                               package_def)

        # fill package.json
        package_lock_def = _generate_package_lock_nodejs_lambda(lambda_name)
        _write_content_to_file(f'{lambda_folder}/'f'{FILE_PACKAGE_LOCK}',
                               package_lock_def)

        # fill deployment_resources.json
        role_def = _generate_lambda_role_config(
            LAMBDA_ROLE_NAME_PATTERN.format(lambda_name))
        _write_content_to_file(f'{lambda_folder}/{FILE_DEPLOYMENT_RESOURCES}',
                               role_def)

        # fill lambda_config.json
        lambda_def = _generate_nodejs_node_lambda_config(
            lambda_name,
            f'{FOLDER_LAMBDAS}/{lambda_name}')
        _write_content_to_file(f'{lambda_folder}/{FILE_LAMBDA_CONFIG}',
                               lambda_def)
        _LOG.info(f'Lambda {lambda_name} created')


LAMBDAS_PROCESSORS = {
    PROJECT_JAVA: _generate_java_lambdas,
    PROJECT_NODEJS: _generate_nodejs_lambdas,
    PROJECT_PYTHON: _generate_python_lambdas,
}



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from syndicate.core.constants import (API_GATEWAY_TYPE, CLOUD_WATCH_ALARM_TYPE,
                                      CLOUD_WATCH_RULE_TYPE, COGNITO_TYPE,
                                      DYNAMO_TABLE_TYPE, EBS_TYPE,
                                      EC2_INSTANCE_TYPE, IAM_POLICY,
                                      IAM_ROLE, KINESIS_STREAM_TYPE,
                                      LAMBDA_TYPE, S3_BUCKET_TYPE,
                                      SNS_PLATFORM_APPLICATION_TYPE,
                                      SNS_TOPIC_TYPE,
                                      SQS_QUEUE_TYPE, STATE_ACTIVITY_TYPE,
                                      STEP_FUNCTION_TYPE, LAMBDA_LAYER_TYPE)


class ProcessorFacade:

    def __init__(self, resources_provider) -> None:
        self.resources_provider = resources_provider

    def create_handlers(self):
        return {
            IAM_POLICY:
                self.resources_provider.iam().create_policies,
            IAM_ROLE:
                self.resources_provider.iam().create_roles,
            LAMBDA_TYPE:
                self.resources_provider.lambda_resource().create_lambda,
            LAMBDA_LAYER_TYPE:
                self.resources_provider.lambda_resource().create_lambda_layer,
            DYNAMO_TABLE_TYPE:
                self.resources_provider.dynamodb().create_tables_by_10,
            CLOUD_WATCH_RULE_TYPE:
                self.resources_provider.cw().create_cloud_watch_rule,
            S3_BUCKET_TYPE:
                self.resources_provider.s3().create_s3_bucket,
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().create_api_gateway,
            COGNITO_TYPE:
                self.resources_provider.cognito().create_cognito_identity_pool,
            SNS_TOPIC_TYPE:
                self.resources_provider.sns().create_sns_topic,
            SNS_PLATFORM_APPLICATION_TYPE:
                self.resources_provider.sns().create_sns_application,
            SQS_QUEUE_TYPE:
                self.resources_provider.sqs().create_sqs_queue,
            CLOUD_WATCH_ALARM_TYPE:
                self.resources_provider.cw_alarm().create_alarm,
            EBS_TYPE:
                self.resources_provider.ebs().create_ebs,
            STEP_FUNCTION_TYPE:
                self.resources_provider.step_functions().create_state_machine,
            STATE_ACTIVITY_TYPE:
                self.resources_provider.step_functions().create_activities,
            KINESIS_STREAM_TYPE:
                self.resources_provider.kinesis().create_kinesis_stream,
            EC2_INSTANCE_TYPE:
                self.resources_provider.ec2().create_ec2
        }

    def describe_handlers(self):
        return {
            IAM_POLICY:
                self.resources_provider.iam().describe_policy,
            IAM_ROLE:
                self.resources_provider.iam().describe_role,
            LAMBDA_TYPE:
                self.resources_provider.lambda_resource().describe_lambda,
            DYNAMO_TABLE_TYPE:
                self.resources_provider.dynamodb().describe_table,
            CLOUD_WATCH_RULE_TYPE:
                self.resources_provider.cw().describe_rule_from_meta,
            S3_BUCKET_TYPE:
                self.resources_provider.s3().describe_bucket,
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().describe_api_resources,
            COGNITO_TYPE:
                self.resources_provider.cognito().describe_cognito_pool,
            SNS_TOPIC_TYPE:
                self.resources_provider.sns().describe_sns_from_meta,
            SNS_PLATFORM_APPLICATION_TYPE:
                self.resources_provider.sns().describe_sns_application_from_meta,
            SQS_QUEUE_TYPE:
                self.resources_provider.sqs().describe_queue_from_meta,
            CLOUD_WATCH_ALARM_TYPE:
                self.resources_provider.cw_alarm().describe_alarm,
            EBS_TYPE:
                self.resources_provider.ebs().describe_ebs,
            STEP_FUNCTION_TYPE:
                self.resources_provider.step_functions().describe_step_function,
            STATE_ACTIVITY_TYPE:
                self.resources_provider.step_functions().describe_activity,
            KINESIS_STREAM_TYPE:
                self.resources_provider.kinesis().describe_kinesis_stream,
        }

    def remove_handlers(self):
        return {
            CLOUD_WATCH_ALARM_TYPE:
                self.resources_provider.cw_alarm().remove_alarms,
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().remove_api_gateways,
            CLOUD_WATCH_RULE_TYPE:
                self.resources_provider.cw().remove_cloud_watch_rules,
            COGNITO_TYPE:
                self.resources_provider.cognito().remove_cognito_identity_pools,
            DYNAMO_TABLE_TYPE:
                self.resources_provider.dynamodb().remove_dynamodb_tables,
            EBS_TYPE:
                self.resources_provider.ebs().remove_ebs_apps,
            EC2_INSTANCE_TYPE:
                self.resources_provider.ec2().remove_ec2_instances,
            IAM_ROLE:
                self.resources_provider.iam().remove_roles,
            IAM_POLICY:
                self.resources_provider.iam().remove_policies,
            KINESIS_STREAM_TYPE:
                self.resources_provider.kinesis().remove_kinesis_streams,
            LAMBDA_LAYER_TYPE:
                self.resources_provider.lambda_resource().remove_lambda_layers,
            LAMBDA_TYPE:
                self.resources_provider.lambda_resource().remove_lambdas,
            S3_BUCKET_TYPE:
                self.resources_provider.s3().remove_buckets,
            SNS_TOPIC_TYPE:
                self.resources_provider.sns().remove_sns_topics,
            SNS_PLATFORM_APPLICATION_TYPE:
                self.resources_provider.sns().remove_sns_application,
            SQS_QUEUE_TYPE:
                self.resources_provider.sqs().remove_queues,
            STEP_FUNCTION_TYPE:
                self.resources_provider.step_functions().remove_state_machines,
            STATE_ACTIVITY_TYPE:
                self.resources_provider.step_functions().remove_activities
        }

    def update_handlers(self):
        return {
            LAMBDA_TYPE:
                self.resources_provider.lambda_resource().update_lambda,
            LAMBDA_LAYER_TYPE:
                self.resources_provider.lambda_resource().update_lambda_layer
        }

    def resource_configuration_processor(self):
        return {
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().api_gateway_update_processor
        }

    def resource_identifier(self):
        return {
            API_GATEWAY_TYPE:
                self.resources_provider.api_gw().api_resource_identifier,
            COGNITO_TYPE:
                self.resources_provider.cognito().cognito_resource_identifier
        }

    def mapping_applier(self):
        return {
            IAM_ROLE: self.resources_provider.iam().apply_trusted_to_role,
            IAM_POLICY: self.resources_provider.iam().apply_policy_content
        }
